#!/usr/bin/lua

local ffi = require 'ffi'
local C = ffi.C
local lpeg = require 'lpeg'
local event = ...

ffi.cdef[[
char *getenv(const char *name);
int unlink(const char *pathname);
int rename(const char *oldpath, const char *newpath);
]]

local params = { }
local started = false
local plist = {
    'interface',
    'subnet',
    'broadcast',
    'ip',
    'router',
    'domain',
    'dns'}

for _, key in ipairs(plist) do
    local v = C.getenv(key)
    if v ~= nil then
        params[key] = ffi.string(v)
    end
end

local function split(s, sep)
    sep = lpeg.P(sep)
    local elem = lpeg.C((1 - sep)^0)
    local p = lpeg.Ct(elem * (sep * elem)^0)
    return lpeg.match(p, s)
end

-- This bit of code was taken from network.lua
local p, e = params, os.execute
if event == 'deconfig' then
    e('/sbin/ifconfig '..p.interface..' 0.0.0.0')
    C.unlink('/etc/resolv.conf')
elseif event == 'renew' or event == 'bound' then
    local netmask = p.subnet and 'netmask '..p.subnet or ''
    local broadcast = p.broadcast and 'broadcast '..p.broadcast
        or 'broadcast +'
    e('/sbin/ifconfig '..p.interface..' '..p.ip..' '
        ..netmask..' '..broadcast)
    if p.router then
        while e('/sbin/route del default gw 0.0.0.0 '
                ..'dev '..p.interface .. ' 2>/dev/null') == 0 do
            -- nothing
        end
        local metric = 0
        for _,router in ipairs(split(p.router, ' ')) do
            e('/sbin/route add default gw '..router..' dev '..p.interface
                ..' metric '..metric)
                metric = metric + 1
        end
    end
    local rc = io.open('/etc/resolv.conf.new', 'w')
    if p.domain then rc:write('domain '..p.domain..'\n') end
    for _,ns in ipairs(split(p.dns, ' ')) do
        rc:write('nameserver '..ns..'\n')
    end
    rc:close()
    C.rename('/etc/resolv.conf.new', '/etc/resolv.conf')
end
