--
-- AXIS HTML (old) 5.2.4.6 MJPG video CGI reguest
--
-- Method: GET
-- Syntax:
--     http://<camera_ip>/axis-cgi/mjpg/video.cgi
--             [?<parameter>=<value>[&<parameter>=<value>...]]
--
-- Parameter Pairs
-- ---------------------------------------------------------------------
-- [ resolution=<string> ]
--     Values: 640x480, 480x360, 320x240, 240x180, 176x144, 160x120
-- [ camera=<string> ]
--     Values: 1, 2, 3, 4, and quad
-- [ compression=<int> ]
--     Values: 0 - 100 (where 0 = largest image size / best quality)
-- [ fps=<int> ]
--     Values: 0, ... (where 0 = unlimited or [30])
--
-- FIXME: The remaining CGI API from documentation

local __ = require '__'
local log = require 'log'
local class = require 'class'
local camera = require 'camera'
local net = require 'net_ffi'
local ffi = require 'ffi'
local C = ffi.C

-- GNU C Library 'libc'

ffi.cdef[[
char *strstr(const char *haystack, const char *needle);
char *strerror(int);
size_t strlen(const char *s);
typedef int32_t pid_t;
pid_t getpid(void);
in_addr_t inet_addr(const char *cp);
]]

-- mDNS Networking

ffi.cdef[[
enum {
    DNS_A     = 1,
    DNS_NS    = 2,
    DNS_CNAME = 5,
    DNS_SOA   = 6,
    DNS_PTR   = 12,
    DNS_HINFO = 13,
    DNS_TXT   = 16,
};

typedef struct dns_header {
    uint16_t id;             /* identification number */

    unsigned char rd :1;     /* recursion desired */
    unsigned char tc :1;     /* truncated message */
    unsigned char aa :1;     /* authoritive answer */
    unsigned char opcode :4; /* purpose of message */
    unsigned char qr :1;     /* query/response flag */

    unsigned char rcode :4;  /* response code */
    unsigned char cd :1;     /* checking disabled */
    unsigned char ad :1;     /* authenticated data */
    unsigned char z :1;      /* its z! reserved */
    unsigned char ra :1;     /* recursion available */

    uint16_t qdcount;        /* number of question entries */
    uint16_t ancount;        /* number of answer entries */
    uint16_t nscount;        /* number of authority entries */
    uint16_t arcount;        /* number of resource entries */
} dns_header_t;

typedef struct dns_question {
    uint16_t qtype;
    uint16_t qclass;
} dns_question_t;

typedef struct dns_resource_data {
    uint16_t type;
    uint16_t class;
    unsigned int   ttl;
    uint16_t len;
} dns_resource_data_t;

typedef struct dns_resource_record {
    unsigned char            name[256];
    struct dns_resource_data *resource;
    unsigned char            *data;
} dns_resource_record_t;

typedef struct dns_query {
    unsigned char       *name;
    struct dns_question *question;
} dns_query_t;

in_addr_t inet_addr(const char *cp);
]]

local m = class(camera)

function m:setup(vendor, model, serial, host, port)
    camera:setup(vendor, model, serial, host, port)
    local query = '?' ..
                  'compression=' ..
                  ffi.string(self.camera[0].spec_image_compression) .. '&' ..
                  'resolution=' ..
                  ffi.string(self.camera[0].spec_image_size) .. '&' ..
                  'fps=' ..
                  ffi.string(self.camera[0].spec_frame_rate)
    if ffi.string(self.camera[0].query) ~= query then
        ffi.copy(self.camera[0].query, query)
    end
    ffi.copy(self.camera[0].page, 'axis-cgi/mjpg/video.cgi')
    return self
end

function m:mdns()
    local host = ffi.new('char [16]', '224.0.0.251')
	local i = ffi.new('int')
	local rc = ffi.new('int')
    local on = ffi.new('int [1]')
    local ttl = ffi.new('unsigned char [1]')
    local addr = ffi.new('struct sockaddr_in [1]')
    local mlen = ffi.sizeof('struct sockaddr_in')
	local buf = ffi.new('unsigned char [65536]')
    local qname = ffi.new('unsigned char *')
    local header = ffi.new('struct dns_header [1]')
    local hsize = ffi.sizeof('struct dns_header')
    local question = ffi.new('struct dns_question [1]')
    local qsize = ffi.sizeof('struct dns_question')
    local o = {}
    -- setup send socket
	local sd = ffi.new('int')
    sd = C.socket(C.AF_INET, C.SOCK_DGRAM, 0)
    if sd < 0 then
        io.stderr:write(string.format("axis_camera:mdns(): socket %s",
                                      ffi.string(C.strerror(ffi.errno()))))
        return
    end
    ffi.fill(addr, ffi.sizeof('struct sockaddr_in'), 0)
    addr[0].sin_family = C.AF_INET
    addr[0].sin_port = C.htons(ffi.new('uint16_t', 5353))
    addr[0].sin_addr.s_addr = C.inet_addr(host)
    on[0] = 0
    rc = C.setsockopt(sd, C.IPPROTO_IP, C.IP_MULTICAST_LOOP, on,
                      ffi.sizeof(on))
    if rc < 0 then
        io.stderr:write(string.format("axis_camera:mdns(): IP_MULTICAST_LOOP %s",
                                      ffi.string(C.strerror(ffi.errno()))))
        C.close(sd)
        return
    end
    ttl[0] = 64
    rc = C.setsockopt(sd, C.IPPROTO_IP, C.IP_MULTICAST_TTL, ttl,
                      ffi.sizeof(ttl))
    if rc < 0 then
        io.stderr:write(string.format("axis_camera:mdns(): IP_MULTICAST_TTL %s",
                                      ffi.string(C.strerror(ffi.errno()))))
        C.close(sd)
        return
    end
    -- build DNS header
    header[0].id = C.htons(C.getpid())
    header[0].qdcount = C.htons(ffi.new('uint16_t', 1))
    ffi.copy(buf, header, hsize)
    -- build DNS question
    qname = buf + ffi.sizeof('struct dns_header')
    qname[0] = 0xb
    ffi.copy(qname + 1, "_axis-video\004_tcp\005local\0")
    question[0].qtype = C.htons(C.DNS_PTR)
    question[0].qclass = C.htons(0x8001)
    ffi.copy(qname + C.strlen(qname) + 1, question, qsize)
    -- send DNS query to Multicast Destination
    rc = C.sendto(sd, buf, hsize + C.strlen(qname) + 1 + qsize, 0,
                  ffi.cast('struct sockaddr *', addr), mlen)
    if rc < 0 then
        io.stderr:write(string.format("axis_camera:mdns(): sendto %s",
                                      ffi.string(C.strerror(ffi.errno()))))
        C.close(sd)
        return
    end
    C.close(sd)
    C.usleep(100000)
end

function m:probe()
    self:mdns()
end

function m:discover(arp)
    local err
    local o = {}
    for i = 1, #arp do
        if arp[i].serial:match('00408c') then
            table.insert(o, { vendor = 'axis', ip = arp[i].ip })
        end
    end
    if #o == 0 then
        -- No Axis serial numbers found
        return o
    end
    local list = {}
    local query = {}
    local line = ffi.new('char *')
    local _model = ffi.new('char [33]')
    local _serial = ffi.new('char [33]')
    local _firmware = ffi.new('char [33]')
    local buf = ffi.new('char [512]')
    for i = 1, #o do
        query = "POST /axis-cgi/admin/param.cgi HTTP/1.0\r\n" ..
                "Host: " .. o[i].ip .. "\r\n" ..
                "User-Agent: HBM CX23R\r\n" ..
                "Content-Type: application/x-www-form-urlencoded\r\n" ..
                "Content-Length: 90\r\n" ..
                "Authorization: Basic cm9vdDpyb290\r\n\r\n" ..
                "action=list&group=Brand.ProdNbr," ..
                                  "Properties.System.SerialNumber," ..
                                  "Properties.Firmware.Version"
        ffi.fill(buf, ffi.sizeof(buf), 0)
        if self:info(o[i].ip, query, buf) then
            line = C.strstr(buf, "ProdNbr")
            if line ~= nil then
                ffi.fill(_model, ffi.sizeof(_model), 0)
                C.sscanf(line, "ProdNbr=%s", _model)
                o[i].model = ffi.string(_model)
            end
            line = C.strstr(line, "SerialNumber")
            if line ~= nil then
                ffi.fill(_serial, ffi.sizeof(_serial), 0)
                C.sscanf(line, "SerialNumber=%s", _serial)
                o[i].serial = string.lower(ffi.string(_serial))
            end
            line = C.strstr(line, "Firmware")
            if line ~= nil then
                ffi.fill(_firmware, ffi.sizeof(_firmware), 0)
                C.sscanf(line, "Firmware.Version=%s", _firmware)
                o[i].firmware = ffi.string(_firmware)
            end
            if o[i].model and o[i].serial and o[i].firmware then
                table.insert(list, o[i])
            else
                if not o[i].model then
                    err = string.format("discover: no model from %s camera @ %s",
                               o[i].vender, o[i].ip)
                    io.stderr:write(err)
                end
                if not o[i].serial then
                    err = string.format("discover: no serial from %s camera @ %s",
                               o[i].vender, o[i].ip)
                    io.stderr:write(err)
                end
                if not o[i].firmware then
                    err = string.format("discover: no firmware from %s camera @ %s",
                               o[i].vender, o[i].ip)
                    io.stderr:write(err)
                end
            end
        else
            err = string.format("discover(): could not query %s camera @ %s",
                                o[i].vender, o[i].ip)
            io.stdout:write(err)
        end
    end
    return list
end

function m:snapshot(proxy, file)
    if not file then
        log:error(__"{0}_camera:snapshot missing filename on {1}",
                  self:vendor(), proxy.proxy_hostname)
    end
    local fn = '/tmp/' .. file
    self.fp = C.fopen(fn, 'wb')
    if self.fp ~= nil then
        os.execute('curl -s -k http://' .. self:host() ..
                   '/jpg/image.jpg' ..
                   ' -o ' .. fn)
        os.execute('ln -sf ' .. fn ..
                   ' /usr/share/web/static/images/' .. file)
    else
        log:error(__"{0}_camera:snapshot {1} on {2}", self:vendor(),
                  ffi.string(C.strerror(ffi.errno())), proxy.proxy_hostname)
    end
    return self.fp
end

return m
