local cc_parser = {}

local lpeg = require 'lpeg'

lpeg.setmaxstack(1024)

local P, S, R, V, C, Cc, Cp, Ct =
    lpeg.P, lpeg.S, lpeg.R, lpeg.V, lpeg.C, lpeg.Cc, lpeg.Cp, lpeg.Ct

local SP = S' \t\r\n'
local sp = SP^0

local function Cas(type, patt)
    return sp * Cp() * Ct(Cc(type) * patt) * Cp() / function(first, cap, last)
                                                        cap.first = first
                                                        cap.last = last
                                                        return cap
                                                    end
end

local function CasG(type, patt)
    return Cas(type, patt) / function(cap)
                                 if #cap == 2 then
                                     return cap[2]
                                 end
                                 return cap
                             end
end

local function spC(p) return sp * C(p) end
local function spP(p) return sp * P(p) end
local function spV(p) return sp * V(p) end

local g = P{'only_expr',
    only_expr = V'expr' * SP^0 * -P(1),
    chunk = Cas('chunk', (V'stmt' * ((spP';' + SP^1) * V'stmt')^0)^-1),
    stmt = Cas('lassign', P'local' * V'name' * spP'=' * V'expr') +
           Cas('assign', V'name' * spP'=' * V'expr'),

    expr = V'lor',
    lor = CasG('lor', V'land' * (V'lorop' * V'land')^0),
    land = CasG('land', V'eq' * (V'landop' * V'eq')^0),
    eq = CasG('eq', V'comp' * (V'eqop' * V'comp')^0),
    comp = CasG('comp', V'sum' * (V'compop' * V'sum')^0),
    sum = CasG('sum', V'product' * (V'sumop' * V'product')^0),
    product = CasG('product', V'exponent' * (V'productop' * V'exponent')^0),
    exponent = CasG('exponent', V'unary' * (V'exponentop' * V'unary')^0),
    unary = Cas('unary', C(S'-+!') * V'unary') + V'value',
    value = CasG('value',
                 V'funcall' + V'message' + V'number' + V'name' + V'symbol' +
                     P'(' * V'expr' * spP')'),

    funcall = Cas('funcall', V'name' * spP'(' * 
                  (V'expr' * (spP',' * V'expr')^0)^-1 * spP')'),

    lorop = Cas('op', C(P'||')),
    landop = Cas('op', C(P'&&')),
    eqop = Cas('op', C(P'==' + P'!=')),
    compop = Cas('op', C(P'>=' + P'<=' + P'>' + P'<')),
    sumop = Cas('op', C(S'+-')),
    productop = Cas('op', C(S'*/%')),
    exponentop = Cas('op', C(S'^')),

    number = Cas('number',
         C((P'0x' * V'hexdigit'^1 * (P'.' * V'hexdigit'^0)^-1) / string.lower)
       + C(S'0123456789.'^1 * (S'eE' * S'+-'^-1 * S'0123456789'^1)^-1)),

    hexdigit = S'0123456789ABCDEFabcdef',
    message = Cas('message', P'0m' * C((V'hexdigit' * V'hexdigit')^1)),

    name = Cas('name', V'barename' + V'quotename'),
    barename = Ct(V'barenamechar'^1) / table.concat,
    barenamechar = P'\\' * C(1) + C(1 - S' \t\r\n+-*/%!~=<>&^?@.,()&|\'\"'),
    quotename = P"'" * Ct(V'quotenamechar'^1) * P"'" / table.concat,
    quotenamechar = P'\\' * C(1) + C(1 - S"'"),
    symbol = Cas('symbol', V'dblquotename'),
    dblquotename = P'"' * Ct(V'dblquotenamechar'^1) * P'"' / table.concat,
    dblquotenamechar = P'\\' * C(1) + C(1 - S'"'),
}

function cc_parser.parse_expression(expr)
    local tree = lpeg.match(g, expr)
    if not tree then
        error(string.format('Expression syntax error in "%s"', expr))
    end
    return tree
end

function cc_parser.getnumber(expr)
    if expr[1] == 'number' then
        return assert(tonumber(expr[2]))
    elseif expr[1] == 'unary' then
        return assert(tonumber(expr[2] .. cc_parser.getnumber(expr[3])))
    else
        error('not a number')
    end
end

return cc_parser
