local setup_spec = require 'setup_spec'

local m = {}

local acc = {}

function acc.from_name(edge)
    return edge.from_node.meta.name
end

function acc.to_name(edge)
    return edge.to_node.meta.name
end

function acc.from_node(edge)
    return edge.compiler:get_node(edge.from_idx)
end

function acc.to_node(edge)
    return edge.compiler:get_node(edge.to_idx)
end

function acc.buf(edge)
    return string.format('bufs[%s]', edge.idx)
end

function acc.data_type(edge)
    -- data_type is special because an edge (ring buffer) can have
    -- a different data_type than the upstream node
    return rawget(edge, 'data_type') or edge.from_node.data_type
end

local fn = {}

function fn.dump(self)
    return setup_spec:item_export(self)
end

local edge_mt = {
    __index = function(table, key)
        if acc[key] then
            return acc[key](table)
        end
        return fn[key]
    end,
}

function m:new(compiler, from_node, to_node)
    fn.compiler = compiler
    from_node = compiler:get_node(from_node)
    to_node = compiler:get_node(to_node)
    local ci = #compiler.edges + 1
    local edge = {
        idx = ci,
        from_idx = from_node.idx,
        to_idx = to_node.idx,
        is_logical = from_node.meta.is_logical
    }
    setmetatable(edge, edge_mt)
    compiler.edges[ci] = edge
    return edge
end

return m
