--! \file

local lpeg = require "lpeg"

--! A parser for HTTP headers.
--? class http_parser
local m = {}
--? alias m http_parser

local      P,      R,      S,      V
    = lpeg.P, lpeg.R, lpeg.S, lpeg.V
local      C,      Carg,      Cb,      Cc,      Cf,      Cg,      Cmt,      Ct
    = lpeg.C, lpeg.Carg, lpeg.Cb, lpeg.Cc, lpeg.Cf, lpeg.Cg, lpeg.Cmt, lpeg.Ct

local whitespace = S(" \t")
local ws = whitespace^1
local ows = whitespace^0
local nonws = (1 - S" \t\r\n")
local crlf = P"\r\n"

local function fix_headers(...)
    local t = {...}
    local r = {}
    for i = 2, #t, 2 do
        r[t[i - 1]:lower():gsub("-", "_")] = t[i]:gsub("[\r\n]", "")
    end
    return r
end

local http_version = Cg("HTTP/1." * S"01", "http_version")

m.headers = P{
    'headers'; -- Start rule
    headers = Cg(V"header"^0 / fix_headers, "headers"),
    header = C(R("az", "AZ", "--")^1) * ":" * ows
        * C((1 - crlf)^0 * (crlf * ws * (1 - crlf))^0) * crlf,
}

m.request = P{
    'request'; -- Start rule
    request = Ct(V"request_line" * m.headers * crlf) * lpeg.Cp(),
    request_line = V"method" * ws * V"uri" * ws * http_version * crlf,
    method = Cg(P"GET" + "HEAD" + "POST" + "PUT", "method"),
    uri = Cg((1 - whitespace)^1, "uri"), -- KLUDGE WRONG
    message_body = Cg(P(1)^0, "message_body"),
}

m.response = P{
    'response'; -- Start rule
    response = Ct(V"response_line" * m.headers * crlf) * lpeg.Cp(),
    response_line = http_version * ws * V"response_code" * ws * V"message"
        * crlf,
    response_code = Cg(R("09")^1, "response_code"),
    message = Cg((1 - S"\r\n")^1, "message"),
}

function m.parse_response(buf) -- KLUDGE add start_pos?
    local http, end_pos = http_parser.response:match(buf)
    if not http then
        return nil
    end

    local clen = http.headers.content_length
    if clen and clen:match("^%d+$") then
        clen = tonumber(clen)
        if buf:len() < end_pos + clen - 1 then
            return nil
        end
        http.content = string.sub(buf, end_pos, end_pos + clen)
        assert(clen == http.content:len())
        end_pos = end_pos + clen
    end
    return http, end_pos
end

return m
