--! \file

local class = require "class"
local data_dumper = require "data_dumper"
local ev = require "ev"
local http_parser = require "http_parser"
local u = require "unix"

--! A simple HTTP server.
--? class http_server server
local m = class(require "server_noffi")
--? alias m http_server

function m.shutdown_timeout(loop, timer, revents)
--print("shutdown_timeout calling close")
    timer.shutdown_w.server.close(loop, timer.shutdown_w)
    timer:stop(loop)
end

function m.http_write_shutdown(loop, w)
    u.shutdown(w.socket, u.SHUT_WR)
    local timer = ev.Timer.new(m.shutdown_timeout, 10, 10)
    timer.shutdown_w = w
    w.shutdown_timer = timer
    timer:start(loop)
end

function m.reader(loop, w, revents)
    local got = u.read(w.socket, 8192)
    if not got or got == "" then
--print("http_server_reader calling close")
        w.server.close(loop, w)
        if w.shutdown_timer then
            w.shutdown_timer:stop(loop)
        end
        return
    end
    w.buf = w.buf .. got

    local http, end_pos = http_parser.request:match(w.buf)
--    print("http_server_reader:\n" .. w.buf)
--    print("http_server_reader:\n" .. data_dumper(http, "", false, 1))
    if not http then
        -- KLUDGE check for crlf crlf, die ???
        return
    end

    local clen = http.headers.content_length
    if clen and clen:match("^%d+$") then
        clen = tonumber(clen)
        if w.buf:len() < end_pos + clen - 1 then
            return
        end
        http.content = string.sub(w.buf, end_pos, end_pos + clen)
        assert(clen == http.content:len())
    end

    w.server.handler(loop, w, http)
end

return m
