local ffi = require 'ffi'
local lj_include = require 'lj_include'

lj_include('linux_if_ether')

ffi.cdef[[
struct ethtool_cmd {
    __u32 cmd;
    __u32 supported;
    __u32 advertising;
    __u16 speed;
    __u8 duplex;
    __u8 port;
    __u8 phy_address;
    __u8 transceiver;
    __u8 autoneg;
    __u8 mdio_support;
    __u32 maxtxpkt;
    __u32 maxrxpkt;
    __u16 speed_hi;
    __u8 eth_tp_mdix;
    __u8 eth_tp_mdix_ctrl;
    __u32 lp_advertising;
    __u32 reserved[2];
};
static __inline__ void ethtool_cmd_speed_set(struct ethtool_cmd *ep,
    __u32 speed)
{
    ep->speed = (__u16) speed;
    ep->speed_hi = (__u16) (speed >> 16);
}
static __inline__ __u32 ethtool_cmd_speed(const struct ethtool_cmd *ep)
{
    return (ep->speed_hi << 16) | ep->speed;
}
struct ethtool_drvinfo {
    __u32 cmd;
    char driver[32];
    char version[32];
    char fw_version[32];
    char bus_info[32];
    char reserved1[32];
    char reserved2[12];
    __u32 n_priv_flags;
    __u32 n_stats;
    __u32 testinfo_len;
    __u32 eedump_len;
    __u32 regdump_len;
};
struct ethtool_wolinfo {
    __u32 cmd;
    __u32 supported;
    __u32 wolopts;
    __u8 sopass[6];
};
struct ethtool_value {
    __u32 cmd;
    __u32 data;
};
struct ethtool_regs {
    __u32 cmd;
    __u32 version;
    __u32 len;
    __u8 data[0];
};
struct ethtool_eeprom {
    __u32 cmd;
    __u32 magic;
    __u32 offset;
    __u32 len;
    __u8 data[0];
};
struct ethtool_eee {
    __u32 cmd;
    __u32 supported;
    __u32 advertised;
    __u32 lp_advertised;
    __u32 eee_active;
    __u32 eee_enabled;
    __u32 tx_lpi_enabled;
    __u32 tx_lpi_timer;
    __u32 reserved[2];
};
struct ethtool_modinfo {
    __u32 cmd;
    __u32 type;
    __u32 eeprom_len;
    __u32 reserved[8];
};
struct ethtool_coalesce {
    __u32 cmd;
    __u32 rx_coalesce_usecs;
    __u32 rx_max_coalesced_frames;
    __u32 rx_coalesce_usecs_irq;
    __u32 rx_max_coalesced_frames_irq;
    __u32 tx_coalesce_usecs;
    __u32 tx_max_coalesced_frames;
    __u32 tx_coalesce_usecs_irq;
    __u32 tx_max_coalesced_frames_irq;
    __u32 stats_block_coalesce_usecs;
    __u32 use_adaptive_rx_coalesce;
    __u32 use_adaptive_tx_coalesce;
    __u32 pkt_rate_low;
    __u32 rx_coalesce_usecs_low;
    __u32 rx_max_coalesced_frames_low;
    __u32 tx_coalesce_usecs_low;
    __u32 tx_max_coalesced_frames_low;
    __u32 pkt_rate_high;
    __u32 rx_coalesce_usecs_high;
    __u32 rx_max_coalesced_frames_high;
    __u32 tx_coalesce_usecs_high;
    __u32 tx_max_coalesced_frames_high;
    __u32 rate_sample_interval;
};
struct ethtool_ringparam {
    __u32 cmd;
    __u32 rx_max_pending;
    __u32 rx_mini_max_pending;
    __u32 rx_jumbo_max_pending;
    __u32 tx_max_pending;
    __u32 rx_pending;
    __u32 rx_mini_pending;
    __u32 rx_jumbo_pending;
    __u32 tx_pending;
};
struct ethtool_channels {
    __u32 cmd;
    __u32 max_rx;
    __u32 max_tx;
    __u32 max_other;
    __u32 max_combined;
    __u32 rx_count;
    __u32 tx_count;
    __u32 other_count;
    __u32 combined_count;
};
struct ethtool_pauseparam {
    __u32 cmd;
    __u32 autoneg;
    __u32 rx_pause;
    __u32 tx_pause;
};
enum ethtool_stringset {
    ETH_SS_TEST = 0,
    ETH_SS_STATS,
    ETH_SS_PRIV_FLAGS,
    ETH_SS_NTUPLE_FILTERS,
    ETH_SS_FEATURES,
};
struct ethtool_gstrings {
    __u32 cmd;
    __u32 string_set;
    __u32 len;
    __u8 data[0];
};
struct ethtool_sset_info {
    __u32 cmd;
    __u32 reserved;
    __u64 sset_mask;
    __u32 data[0];
};
enum ethtool_test_flags {
    ETH_TEST_FL_OFFLINE = (1 << 0),
    ETH_TEST_FL_FAILED = (1 << 1),
    ETH_TEST_FL_EXTERNAL_LB = (1 << 2),
    ETH_TEST_FL_EXTERNAL_LB_DONE = (1 << 3),
};
struct ethtool_test {
    __u32 cmd;
    __u32 flags;
    __u32 reserved;
    __u32 len;
    __u64 data[0];
};
struct ethtool_stats {
    __u32 cmd;
    __u32 n_stats;
    __u64 data[0];
};
struct ethtool_perm_addr {
    __u32 cmd;
    __u32 size;
    __u8 data[0];
};
enum ethtool_flags {
    ETH_FLAG_TXVLAN = (1 << 7),
    ETH_FLAG_RXVLAN = (1 << 8),
    ETH_FLAG_LRO = (1 << 15),
    ETH_FLAG_NTUPLE = (1 << 27),
    ETH_FLAG_RXHASH = (1 << 28),
};
struct ethtool_tcpip4_spec {
    __be32 ip4src;
    __be32 ip4dst;
    __be16 psrc;
    __be16 pdst;
    __u8 tos;
};
struct ethtool_ah_espip4_spec {
    __be32 ip4src;
    __be32 ip4dst;
    __be32 spi;
    __u8 tos;
};
struct ethtool_usrip4_spec {
    __be32 ip4src;
    __be32 ip4dst;
    __be32 l4_4_bytes;
    __u8 tos;
    __u8 ip_ver;
    __u8 proto;
};
union ethtool_flow_union {
    struct ethtool_tcpip4_spec tcp_ip4_spec;
    struct ethtool_tcpip4_spec udp_ip4_spec;
    struct ethtool_tcpip4_spec sctp_ip4_spec;
    struct ethtool_ah_espip4_spec ah_ip4_spec;
    struct ethtool_ah_espip4_spec esp_ip4_spec;
    struct ethtool_usrip4_spec usr_ip4_spec;
    struct ethhdr ether_spec;
    __u8 hdata[52];
};
struct ethtool_flow_ext {
    __u8 padding[2];
    unsigned char h_dest[6];
    __be16 vlan_etype;
    __be16 vlan_tci;
    __be32 data[2];
};
struct ethtool_rx_flow_spec {
    __u32 flow_type;
    union ethtool_flow_union h_u;
    struct ethtool_flow_ext h_ext;
    union ethtool_flow_union m_u;
    struct ethtool_flow_ext m_ext;
    __u64 ring_cookie;
    __u32 location;
};
struct ethtool_rxnfc {
    __u32 cmd;
    __u32 flow_type;
    __u64 data;
    struct ethtool_rx_flow_spec fs;
    __u32 rule_cnt;
    __u32 rule_locs[0];
};
struct ethtool_rxfh_indir {
    __u32 cmd;
    __u32 size;
    __u32 ring_index[0];
};
struct ethtool_rx_ntuple_flow_spec {
    __u32 flow_type;
    union {
	struct ethtool_tcpip4_spec tcp_ip4_spec;
	struct ethtool_tcpip4_spec udp_ip4_spec;
	struct ethtool_tcpip4_spec sctp_ip4_spec;
	struct ethtool_ah_espip4_spec ah_ip4_spec;
	struct ethtool_ah_espip4_spec esp_ip4_spec;
	struct ethtool_usrip4_spec usr_ip4_spec;
	struct ethhdr ether_spec;
	__u8 hdata[72];
    } h_u, m_u;
    __u16 vlan_tag;
    __u16 vlan_tag_mask;
    __u64 data;
    __u64 data_mask;
    __s32 action;
};
struct ethtool_rx_ntuple {
    __u32 cmd;
    struct ethtool_rx_ntuple_flow_spec fs;
};
enum ethtool_flash_op_type {
    ETHTOOL_FLASH_ALL_REGIONS = 0,
};
struct ethtool_flash {
    __u32 cmd;
    __u32 region;
    char data[128];
};
struct ethtool_dump {
    __u32 cmd;
    __u32 version;
    __u32 flag;
    __u32 len;
    __u8 data[0];
};
struct ethtool_get_features_block {
    __u32 available;
    __u32 requested;
    __u32 active;
    __u32 never_changed;
};
struct ethtool_gfeatures {
    __u32 cmd;
    __u32 size;
    struct ethtool_get_features_block features[0];
};
struct ethtool_set_features_block {
    __u32 valid;
    __u32 requested;
};
struct ethtool_sfeatures {
    __u32 cmd;
    __u32 size;
    struct ethtool_set_features_block features[0];
};
struct ethtool_ts_info {
    __u32 cmd;
    __u32 so_timestamping;
    __s32 phc_index;
    __u32 tx_types;
    __u32 tx_reserved[3];
    __u32 rx_filters;
    __u32 rx_reserved[3];
};
enum ethtool_sfeatures_retval_bits {
    ETHTOOL_F_UNSUPPORTED__BIT,
    ETHTOOL_F_WISH__BIT,
    ETHTOOL_F_COMPAT__BIT,
};
enum ethtool_reset_flags {
    ETH_RESET_MGMT = 1 << 0,
    ETH_RESET_IRQ = 1 << 1,
    ETH_RESET_DMA = 1 << 2,
    ETH_RESET_FILTER = 1 << 3,
    ETH_RESET_OFFLOAD = 1 << 4,
    ETH_RESET_MAC = 1 << 5,
    ETH_RESET_PHY = 1 << 6,
    ETH_RESET_RAM = 1 << 7,
    ETH_RESET_DEDICATED = 0x0000ffff,
    ETH_RESET_ALL = 0xffffffff,
};
]]

local defs = {
    ETH_MDIO_SUPPORTS_C22 = '1',
    ETH_MDIO_SUPPORTS_C45 = '2',
    ETHTOOL_FWVERS_LEN = '32',
    ETHTOOL_BUSINFO_LEN = '32',
    SOPASS_MAX = '6',
    ETH_GSTRING_LEN = '32',
    ETH_RX_NFC_IP4 = '1',
    ETHTOOL_FLASH_MAX_FILENAME = '128',
    ETH_FW_DUMP_DISABLE = '0',
    ETHTOOL_GSET = '0x00000001',
    ETHTOOL_SSET = '0x00000002',
    ETHTOOL_GDRVINFO = '0x00000003',
    ETHTOOL_GREGS = '0x00000004',
    ETHTOOL_GWOL = '0x00000005',
    ETHTOOL_SWOL = '0x00000006',
    ETHTOOL_GMSGLVL = '0x00000007',
    ETHTOOL_SMSGLVL = '0x00000008',
    ETHTOOL_NWAY_RST = '0x00000009',
    ETHTOOL_GLINK = '0x0000000a',
    ETHTOOL_GEEPROM = '0x0000000b',
    ETHTOOL_SEEPROM = '0x0000000c',
    ETHTOOL_GCOALESCE = '0x0000000e',
    ETHTOOL_SCOALESCE = '0x0000000f',
    ETHTOOL_GRINGPARAM = '0x00000010',
    ETHTOOL_SRINGPARAM = '0x00000011',
    ETHTOOL_GPAUSEPARAM = '0x00000012',
    ETHTOOL_SPAUSEPARAM = '0x00000013',
    ETHTOOL_GRXCSUM = '0x00000014',
    ETHTOOL_SRXCSUM = '0x00000015',
    ETHTOOL_GTXCSUM = '0x00000016',
    ETHTOOL_STXCSUM = '0x00000017',
    ETHTOOL_GSG = '0x00000018',
    ETHTOOL_SSG = '0x00000019',
    ETHTOOL_TEST = '0x0000001a',
    ETHTOOL_GSTRINGS = '0x0000001b',
    ETHTOOL_PHYS_ID = '0x0000001c',
    ETHTOOL_GSTATS = '0x0000001d',
    ETHTOOL_GTSO = '0x0000001e',
    ETHTOOL_STSO = '0x0000001f',
    ETHTOOL_GPERMADDR = '0x00000020',
    ETHTOOL_GUFO = '0x00000021',
    ETHTOOL_SUFO = '0x00000022',
    ETHTOOL_GGSO = '0x00000023',
    ETHTOOL_SGSO = '0x00000024',
    ETHTOOL_GFLAGS = '0x00000025',
    ETHTOOL_SFLAGS = '0x00000026',
    ETHTOOL_GPFLAGS = '0x00000027',
    ETHTOOL_SPFLAGS = '0x00000028',
    ETHTOOL_GRXFH = '0x00000029',
    ETHTOOL_SRXFH = '0x0000002a',
    ETHTOOL_GGRO = '0x0000002b',
    ETHTOOL_SGRO = '0x0000002c',
    ETHTOOL_GRXRINGS = '0x0000002d',
    ETHTOOL_GRXCLSRLCNT = '0x0000002e',
    ETHTOOL_GRXCLSRULE = '0x0000002f',
    ETHTOOL_GRXCLSRLALL = '0x00000030',
    ETHTOOL_SRXCLSRLDEL = '0x00000031',
    ETHTOOL_SRXCLSRLINS = '0x00000032',
    ETHTOOL_FLASHDEV = '0x00000033',
    ETHTOOL_RESET = '0x00000034',
    ETHTOOL_SRXNTUPLE = '0x00000035',
    ETHTOOL_GRXNTUPLE = '0x00000036',
    ETHTOOL_GSSET_INFO = '0x00000037',
    ETHTOOL_GRXFHINDIR = '0x00000038',
    ETHTOOL_SRXFHINDIR = '0x00000039',
    ETHTOOL_GFEATURES = '0x0000003a',
    ETHTOOL_SFEATURES = '0x0000003b',
    ETHTOOL_GCHANNELS = '0x0000003c',
    ETHTOOL_SCHANNELS = '0x0000003d',
    ETHTOOL_SET_DUMP = '0x0000003e',
    ETHTOOL_GET_DUMP_FLAG = '0x0000003f',
    ETHTOOL_GET_DUMP_DATA = '0x00000040',
    ETHTOOL_GET_TS_INFO = '0x00000041',
    ETHTOOL_GMODULEINFO = '0x00000042',
    ETHTOOL_GMODULEEEPROM = '0x00000043',
    ETHTOOL_GEEE = '0x00000044',
    ETHTOOL_SEEE = '0x00000045',
    SPARC_ETH_GSET = '0x00000001',
    SPARC_ETH_SSET = '0x00000002',
    SPEED_10 = '10',
    SPEED_100 = '100',
    SPEED_1000 = '1000',
    SPEED_2500 = '2500',
    SPEED_10000 = '10000',
    DUPLEX_HALF = '0x00',
    DUPLEX_FULL = '0x01',
    DUPLEX_UNKNOWN = '0xff',
    PORT_TP = '0x00',
    PORT_AUI = '0x01',
    PORT_MII = '0x02',
    PORT_FIBRE = '0x03',
    PORT_BNC = '0x04',
    PORT_DA = '0x05',
    PORT_NONE = '0xef',
    PORT_OTHER = '0xff',
    XCVR_INTERNAL = '0x00',
    XCVR_EXTERNAL = '0x01',
    XCVR_DUMMY1 = '0x02',
    XCVR_DUMMY2 = '0x03',
    XCVR_DUMMY3 = '0x04',
    AUTONEG_DISABLE = '0x00',
    AUTONEG_ENABLE = '0x01',
    ETH_TP_MDI_INVALID = '0x00',
    ETH_TP_MDI = '0x01',
    ETH_TP_MDI_X = '0x02',
    ETH_TP_MDI_AUTO = '0x03',
    TCP_V4_FLOW = '0x01',
    UDP_V4_FLOW = '0x02',
    SCTP_V4_FLOW = '0x03',
    AH_ESP_V4_FLOW = '0x04',
    TCP_V6_FLOW = '0x05',
    UDP_V6_FLOW = '0x06',
    SCTP_V6_FLOW = '0x07',
    AH_ESP_V6_FLOW = '0x08',
    AH_V4_FLOW = '0x09',
    ESP_V4_FLOW = '0x0a',
    AH_V6_FLOW = '0x0b',
    ESP_V6_FLOW = '0x0c',
    IP_USER_FLOW = '0x0d',
    IPV4_FLOW = '0x10',
    IPV6_FLOW = '0x11',
    ETHER_FLOW = '0x12',
    FLOW_EXT = '0x80000000',
    FLOW_MAC_EXT = '0x40000000',

    RX_CLS_LOC_SPECIAL = '0x80000000',
    RX_CLS_LOC_ANY = '0xffffffff',
    RX_CLS_LOC_FIRST = '0xfffffffe',
    RX_CLS_LOC_LAST = '0xfffffffd',
    ETH_MODULE_SFF_8079 = '0x1',
    ETH_MODULE_SFF_8079_LEN = '256',
    ETH_MODULE_SFF_8472 = '0x2',
    ETH_MODULE_SFF_8472_LEN = '512',
    ETH_RESET_SHARED_SHIFT = '16',
}
local format, cdef, C = string.format, ffi.cdef, ffi.C
for sym, val in pairs(defs) do
    if not pcall(function() return C[sym] end) then
        cdef(format("enum {%s = %s};", sym, val))
    end
end

