-- Wrap the whole works so uncaught run time stuff make it into the log
local _module_ok, _module_err_msg = pcall(function()

require('lj_include'){
    'string',
    'sys/types',
    'sys/stat',
    'fcntl',
    'unistd',
    'signal',
}

local armada = require 'armada'
local ev = require 'ev'
local log = require 'log'

log:open('mx-proxy', log.USER)
local ok, mx = pcall(require, 'mx.mx')
if not ok then
    log:log(log.ERR, mx)
    os.exit(1)
end

local channel_proxy = require 'channel_proxy'
local ffi = require 'ffi'
local C = ffi.C
local http_server = require 'http_server'
local cjson = require 'cjson'
local rational = require 'rational'
local netdb = require 'network_db'
local ndb = netdb:new()
local underscore = require 'underscore'
local config = require 'system_config'
local sys_notify = require 'sys_notify'

config.allow_teds = config.allow_teds or false

local mxDeviceProxy = class('mxDeviceProxy', channel_proxy)
local mxChannelProxy = class('mxChannelProxy', channel_proxy)

local mypid = string.format("%.0f", tonumber(C.getpid()))

function mxDeviceProxy:create_channel_proxy(i, is_teds, state)
    local ok, p =
        pcall(mxChannelProxy.new, mxChannelProxy, self, i, is_teds, state)
    if not ok then
        error(p)
        return
    end

    local proxy_name = string.format('connector%d.%s', i, self.name)
    p.proxy_name = proxy_name
    p.module_hostname = self.module_hostname
    p.module_ip = self.module_ip
    p.channel_listen_ip = self.channel_listen_ip
    p.channel_setup_ip = self.channel_setup_ip
    p.proxy_hostname = p.proxy_name
    p.proxy_address = self.module_ip

    local http = http_server:new{
        ev_loop = ev.Loop.default,
        handler = p.http_handler,
        add_to_watchers = {
            channel_proxy = p,
        }
    }
    p.http = http
    local http_w = http:listen{address = p.proxy_address}
    p.http_w = http_w
    p.proxy_port = http_w.port

    local http_local = http_server:new{
        ev_loop = ev.Loop.default,
        handler = p.http_handler,
        add_to_watchers = {
            channel_proxy = p,
        }
    }
    local unix_srv = '/tmp/proxy/' .. proxy_name
    local http_local_w = http:listen{path = unix_srv}
    p.proxy_path = http_local_w.path

    local ch_list = p:available_channels(p.proxy_hostname)
    p.old_available_channels = cjson.encode(ch_list)
    p.comet_sinks = {}
    p.channel_sink = nil
    local host = p.proxy_hostname
    p.channel_name = string.format("@%s", host)
    self.rrs[#self.rrs + 1] =
        {'_armada-connectors', 'PTR', host}
    self.rrs[#self.rrs + 1] =
        {host, 'SRV', '0 0 ' .. http_w.port .. ' ' .. self.module_hostname}
    self.rrs[#self.rrs + 1] =
        {host, 'TXT', 'type: ' .. p.proxy_type}
    self.rrs[#self.rrs + 1] =
        {host, 'TXT', 'available_channels: ' .. p.old_available_channels}

    -- Default the sample rate for the channel to 10Hz
    self.mx.control:set_channel_sample_rate(i, 10)

    -- Default TEDS to off (unless overridden by system config
    self.mx.control:set_channel_teds_enable(i, config.allow_teds)

    return p
end

-- This function is used to process the install/remove TEDS events from the
-- JET connection.
function mxDeviceProxy:process_teds_events()
    -- Only do this if there are event to process AND we are not already
    -- processing an event.
    if #self.teds_events > 0 and not self.processing_teds_event then
        local te = self.teds_events[1]
        table.remove(self.teds_events, 1)
        self.processing_teds_event = true
        self:check_teds(te.ref, te.status, function(r)
                self.processing_teds_event = false
                self:process_teds_events()
            end)
    end
end

function mxDeviceProxy:complete_initialization()
    self.proxy_type = 'mxDeviceProxy'
    self.short_name = self.mx.control.devinfo.name:lower()
    self.name = self.short_name .. '-' ..  string.sub(string.format('%012x',
        self.mx.control.devinfo.serial_number), -6)

    -- Set up the proxies for each of the connectors.  Note, making these
    -- channels available to the system is deferred until the unit has sync
    ev.Signal.new(function() end, C.SIGPIPE):start(ev.Loop.default)

    for _, sig in ipairs{C.SIGINT, C.SIGTERM, C.SIGQUIT} do
        ev.Signal.new(function()
                log:log(log.NOTICE,
                    string.format("Exiting as a result of signal %s", sig))
                os.exit(sig)
            end, sig):start(ev.Loop.default)
    end

    local module_hostname = armada.network:GetModuleHostname()
    local module_ip = armada.network:GetModuleIP()
    local channel_listen_ip = armada.network:GetModuleLinkLocalIP()
    local channel_setup_ip = channel_listen_ip:gsub("%%.*$", "")

    self.module_hostname = module_hostname
    self.module_ip = module_ip
    self.channel_listen_ip = channel_listen_ip
    self.channel_setup_ip = channel_setup_ip

    self.channel_proxies = {}
    self.rrs = {}

    for i = 1, #self.mx.control.channel_info do
        self.channel_proxies[#self.channel_proxies + 1] =
            self:create_channel_proxy(i)
    end

    -- Make the channels available immediately.  See comment below about code
    -- used to defer making the channels available until sync is established.
    self:make_channels_available()

    -- Now we can safely start monitoring the JET stream
    self.mx:start_jet(
        function(sync_info)
            self.sync_info = sync_info
            if self.disregard_sync_status_changes then return end
            if self.start_test_setup_id then
                local msg = "Synchonization status change while running test"
                self:_send_all_channels_error(msg,
                    self.saved_start_channel_list, true)
                log:log(log.ERR, msg)
                for ref, info in pairs(self.stream_start_info) do
                    self:_do_channel_cleanup(ref)
                end

                self:do_post_test_teds()
            end
        end,

        -- Callback for TEDS events
        -- We can't simply process these as the come in because that may result
        -- in multiple install/remove events getting processed in parallel and
        -- that confuses the *&^@# out of an MX device
        function(ref, status)
            if not config.allow_teds then
--[[
                log:log(log.NOTICE, string.format(
                    "TEDS %s event for %s with TEDS disabled",
                    status and 'install' or 'remove', ref))
]]
                return
            elseif self.start_test_setup_id then
                self.deferred_teds_events[ref] = status
                return
            end
            self.teds_events[#self.teds_events + 1] = {ref=ref, status=status}
            self:process_teds_events()
        end)
end

-- Function used to make all of the channels for the device available to the
-- system.  Note, this is deferred until the unit indicates (via JET) that it
-- is synchronized.
function mxDeviceProxy:make_channels_available()
    -- Connector channels
    armada.network:ModifyRRs({}, self.rrs)

    -- Device level channels
    channel_proxy.initialize(self, self.name, true)

    -- Done
    self.rrs = {}
end

-- Function used to add/remove TEDS channels as they come/go.
function mxDeviceProxy:check_teds(ref, status, cb)
    local chan_num = ref:match('_Connector(%d+)$')
    if not chan_num then return end
    chan_num = tonumber(chan_num)
    if not chan_num then return end
    if not self.channel_proxies[chan_num] then return end

    -- If this is a remove event AND the proxy is NOT a proxy for a TEDS
    -- sensor on the connector, just return.
    if not status or status == 'remove' then
        local p = self.channel_proxies[chan_num]
        if p.is_teds then
            -- Remove the existing available channel information
            local rrs = {
                {
                    p.proxy_hostname,
                    'TXT',
                    'available_channels: ' .. p.old_available_channels,
                }
            }
            armada.network:ModifyRRs(rrs, {})

            -- Update the available channel info to show no TEDS sensor
            p.is_teds = nil
            local new_ac = cjson.decode(p.old_available_channels)
            new_ac[1].sensor = {
                use_teds = config.allow_teds and '2' or '0',
                status = 'not connected',
            }
--[[ Leave name as is
            new_ac[1].name = string.format('%s_ch_%d',
                p.mxdev.name:gsub('-', '_'), chan_num)
]]
            p.old_available_channels = cjson.encode(new_ac)
            local rrs = {
                {
                    p.proxy_hostname,
                    'TXT',
                    'available_channels: ' .. p.old_available_channels,
                }
            }
            armada.network:ModifyRRs({}, rrs)
        end
        if cb then cb(ref) end
        return
    end

    -- Remove the existing channel proxy for this connector
    self.channel_proxies[chan_num].destroy(self.channel_proxies[chan_num])

    local retry = 0
    ev.Timer.new(function(l, t, e)
            t:stop(l)
            self.mx.control:get_views(
                {'Analog input connector parameters'},
                function()
                    retry = retry + 1
                    local ok, obj = pcall(self.create_channel_proxy, self,
                        chan_num, true, status)
                    if ok then
                        self.channel_proxies[chan_num] = obj
                        self.channel_proxies[chan_num].is_teds = true
                        armada.network:ModifyRRs({}, self.rrs)
                        self.rrs = {}
                        if cb then cb(ref) end
                    elseif retry > 3 then
                        local sn = sys_notify:new('TEDS Device Error')
                        sn:warning(obj:gsub('.+%.lua:%d+:%s*', ''))
                        if cb then cb(ref) end
--[[
                        -- FAIL - Restore proxy back to normal proxy
                        local ok, obj = pcall(self.create_channel_proxy,
                            self, chan_num, true, 'remove')
                        if ok then
                            self.channel_proxies[chan_num] = obj
                            armada.network:ModifyRRs({}, self.rrs)
                            self.rrs = {}
                        else
                            log:log(log.ERR, string.format(
                                "Unable to restore channel proxy to " ..
                                "plain channel proxy on connetor %d of %s",
                                chan_num, self.proxy_name))
                        end
]]
                    else
                        t:start(l)
                        t:again(l, .1)
                    end
                end)
        end, .25, .1):start(ev.Loop.default)
end

function mxDeviceProxy:initialize(host)
    self.host = host
    self.stream_start_info = {}
    self.teds_events = {}
    self.deferred_teds_events = {}
    self.processing_teds_event = false
    self.mx = mx:new(host, log, ev.Loop.default, function(info)
            if info.result ~= 'error' then
                self:complete_initialization()
            else
                log:log(log.ERR, "MX proxy initialization failed")
                ev.Loop.default:unloop()
            end
        end)
    self.sync_info = { }
    if self.mx then
        return self
    end
end

-- FIXME:  When stacked calibration stuff is available these need to be fixed!
function mxDeviceProxy:available_channels(hostname)
    local n = self.name:gsub('-', '_')
    self.device_channels = {
        default = {
            name = string.format('%s_status', n),
            connector = string.format('error.status@%s', self.proxy_hostname),
            description = 'MX Status Messages',
            data_type = 'message_raw',
            output_units = 'json',
            category = string.format('source/hbm_%s', self.short_name),
            message_rate = 1,
            is_logical = false,
            comet_sinks = {},
        },
        control_channel = {
            name = string.format('%s_control', n),
            connector = string.format('control.status@%s', self.proxy_hostname),
            description = 'MX Unit Control',
            data_type = 'message_raw',
            output_units = '',
            category = string.format('source/hbm_%s', self.short_name),
            message_rate = 1,
            is_logical = false,
            comet_sinks = {},
            valid_controls = {
                change_drd = true,
                led = true,
                reboot = true,
                update_firmware = true,
                get_info = true,
                rezero_channels = true,
            },
        },
        sync_status = {
            name = string.format('%s_sync_status', n),
            connector = string.format('sync.status@%s', self.proxy_hostname),
            description = 'MX Synchronization Status',
            data_type = 'message_raw',
            output_units = 'state',
            category = string.format('source/hbm_%s', self.short_name),
            message_rate = 1,
            is_logical = false,
            comet_sinks = {},
        },
        drd_status = {
            name = string.format('%s_drd_status', n),
            connector = string.format('drd.status@%s', self.proxy_hostname),
            description = 'MX Data Rate Domain',
            data_type = 'message_raw',
            output_units = 'state',
            category = string.format('source/hbm_%s', self.short_name),
            message_rate = 1,
            is_logical = false,
            comet_sinks = {},
            valid_controls = {
                change_drd = true,
            },
        },
        led_status = {
            name = string.format('%s_led_status', n),
            connector = string.format('led.status@%s', self.proxy_hostname),
            description = 'MX Power LED State',
            data_type = 'message_raw',
            output_units = 'state',
            category = string.format('source/hbm_%s', self.short_name),
            message_rate = 1,
            is_logical = false,
            comet_sinks = {},
            valid_controls = {
                led = true,
            },
        },
    }

    -- Add device specific device level controls to the control channel
    local controls = self.mx.control.device.controls and
        self.mx.control.device.controls.device or {}
    for c, f in pairs(controls) do
        self.device_channels.control_channel.valid_controls[c] = true
        self[c] = f
    end

    return {
        self.device_channels.default,
        self.device_channels.control_channel,
        self.device_channels.sync_status,
        self.device_channels.drd_status,
        self.device_channels.led_status,
    }

end

-- Map of DRD "magic" numbers to strings
local drd_map = {
    ['0'] = 'hbm_legacy',
    ['2'] = 'decimal',
    ['3'] = 'binary',
}

-- Function used to send device level channels to the comet stream.  This is
-- run via a timer
function mxDeviceProxy:comet_service()
    -- Error status (default) channel
    if self.status_msg and #self.status_msg > 0 then
        local c = self.device_channels.default
        local msg = cjson.encode(self.status_msg)
        for _, sink in pairs(c.comet_sinks) do
            sink(0, c.connector, msg)
        end
    end

    -- Synchrnoization status
    local c = self.device_channels.sync_status
    for _, sink in pairs(c.comet_sinks) do
        local sync_msg = 'unsynchronized'
        if self.sync_info and self.sync_info.synced then
            sync_msg = self.sync_info.source or 'unsynchronized'
        end
        sink(0, c.connector, sync_msg)
    end

    -- Current Data Rate Domain
    local c = self.device_channels.drd_status
    for _, sink in pairs(c.comet_sinks) do
        local msg = drd_map[self.mx.control.devinfo.current_drd or 5] or
            'unknown'
        sink(0, c.connector, msg)
    end

    -- Current LED Status
    local c = self.device_channels.led_status
    local led_map = {
        normal = 'MX Device Controlled',
        flashing_green = 'Flashing Green',
        flashing_orange = 'Flashing Orange',
        flashing_red = 'Flashing Red',
    }
    for _, sink in pairs(c.comet_sinks) do
        local msg = led_map[self.mx.control.devinfo.current_led_state] or
            'unknown'
        sink(0, c.connector, msg)
    end
end

-- This function is use to supply data to the comet stream for the channels
-- associated with physical connectors on the device.
function mxDeviceProxy:channel_comet_service()
    if not self.mx.stream then return end
    for _, p in ipairs(self.channel_proxies) do
        local sink_list = {}
        local r = p.stream_reference:match("(.+)%.Signal")
        local channel_name = p.channel_name
        sink_list = p.comet_sinks
        for _, sink in pairs(sink_list) do
            local v = self.mx.stream:get_comet_data(p.stream_reference)
            if v then
                sink(0, channel_name, v)
            end
        end
    end
end

function mxDeviceProxy:start_comet_stream(w, chans, sink)
    if not self.comet_service_handler then
        self.comet_service_handler = ev.Timer.new(function(l, t, e)
                self:comet_service()
            end,
            self.device_channels.default.message_rate,
            self.device_channels.default.message_rate)
        self.comet_service_handler:start(ev.Loop.default)
    end
    if not self.comet_channel_service_handler then
        self.comet_channel_service_handler = ev.Timer.new(function(l, t, e)
                self:channel_comet_service()
            end, 1/5, 1/5)
        self.comet_channel_service_handler:start(ev.Loop.default)
    end
    for _, chan in ipairs(chans) do
        for _, v in pairs(self.device_channels) do
            if chan == v.connector then
                v.comet_sinks[w.socket] = sink
                w.cleanup = function()
                    for _, v in pairs(self.device_channels) do
                        v.comet_sinks[w.socket] = nil
                    end
                end
                break
            end
        end
    end
end

function mxDeviceProxy:get_info(params, cb)
    return {
        result = 'success',
        msg = 'MX system information returned',
        current_drd = drd_map[self.mx.control.devinfo.current_drd] or 'unknown',
        serial_number = string.sub(string.format("%012x",
            self.mx.control.devinfo.serial_number), -6),
        firmware_version = self.mx.control.devinfo.fwver or 'unknown',
        expected_firmware_version =
            self.mx.control.devinfo.expected_fwver or 'unknown',
        ip_address = arg[1],
        proxy_host = ndb:get_inside_network_hostname() or 'unknown',
        synchronized = self.sync_info.synced,
    }
end

function mxDeviceProxy:led(params, cb)
    if params[1]:lower() ~= 'flash' then
        self.mx:set_led('normal')
    else
        self.mx:set_led('flashing_orange')
    end
    return {
        result = 'success',
        msg = string.format("LED state set set to %s", params[1] or 'normal')
    }
end

function mxDeviceProxy:update_firmware(params, cb)
    if self.start_test_setup_id then
        return {
            result = 'error',
            msg = 'One or more channels on device are being used in a test',
        }
    end
    self.no_test_start = 'Device is performing user requested firmware update'
    local f = io.popen("ls -t1 /usr/share/quantum-firmware/*.hbg")
    if not f then
        return {
            result = 'error',
            msg = "No MX firmware found on device",
        }
    end
    for fw_image in f:lines() do
        f:close()
        local fifo_name = string.format("/tmp/updater-fifo-%d-%d", C.getpid(),
            math.random(65535))
        local status = C.mkfifo(fifo_name, 436)

        local cmd = string.format('/usr/bin/quantum-updater %s %s >%s &',
            fw_image, self.name, fifo_name)

        -- Tell the stream handling stuff not to exit when heartbeats on the
        -- stream connection stop.
        self.mx.stream.firmware_update_in_progress = true

        os.execute(cmd)

        -- ffi.new for mode because C.open is varargs
        local fd = C.open(fifo_name, C.O_RDWR, ffi.new('mode_t', 0))
        if fd < 0 then
            C.unlink(fifo_name)
            return {
                result = 'error',
                msg = 'Unable to open FIFO to update process'
            }
        end
        local buf = ffi.new('char[8192]')
        local boffset = 0

        -- Read from the FIFO and push everything we get, onto the comet stream
        -- for the control.status channel
        ev.IO.new(function(l, i, e)
                if boffset >= ffi.sizeof(buf) then
                    C.close(fd)
                    i:stop(l)
                    C.unlink(fifo_name)
                    buf = nil
                    local c = self.device_channels.control_channel
                    local msg = string.format(
                        '{"update_progress":{"task":{"name":' ..
                        '"Aborted on out of buffer space","progress":100},' ..
                        '"progress":100}}')
                    for _, sink in pairs(c.comet_sinks) do
                        sink(0, c.connector, msg)
                    end
                    os.exit(-101)
                end
                local amount_read = C.read(fd, buf + boffset,
                    ffi.sizeof(buf) - boffset)
                if amount_read < 0 then
                    if boffset > 0 then
                        local c = self.device_channels.control_channel
                        local msg = string.format('{"update_progress":%s}',
                            ffi.string(buf, boffset))
                        for _, sink in pairs(c.comet_sinks) do
                            sink(0, c.connector, msg)
                        end
                    end
                    C.close(fd)
                    i:stop(l)
                    C.unlink(fifo_name)
                    buf = nil
                    os.exit(-101)
                end
                boffset = boffset + amount_read

                -- Send all complete messages in the buffer to the comet stream
                while boffset > 0 do
                    local offset, e =
                        string.find(ffi.string(buf, boffset), '\n')
                    if not offset and not e then
                        break
                    end
                    local msg = string.format('{"update_progress":%s}',
                        ffi.string(buf, offset - 1))
                    local c = self.device_channels.control_channel
                    for _, sink in pairs(c.comet_sinks) do
                        sink(0, c.connector, msg)
                    end
                    -- If we get the success status (whether or not it is true
                    -- or false) we are done.
                    if msg:find('"success":') then
                        i:stop(l)
                        C.close(fd)
                        C.unlink(fifo_name)
                        buf = nil
                        os.exit(0)
                        return
                    end
                    boffset = boffset - offset
                    if boffset > 0 then 
                        ffi.copy(buf, buf + offset, boffset)
                    elseif boffset < 0 then
                        C.close(fd)
                        break
                    end
                end
            end, fd, ev.READ):start(ev.Loop.default)

        local msg = '{"update_progress":{"task":' ..
            '{"name":"Initializing","progress":0},"progress":0}}'
        local c = self.device_channels.control_channel
        for _, sink in pairs(c.comet_sinks) do
            sink(0, c.connector, msg)
        end

        return {
            result = 'success',
            msg = "Firmware update in progress",
        }
    end
    f:close()
    return {
        result = 'error',
        msg = "Unable to update MX firmware",
    }
end

function mxDeviceProxy:reboot(params, cb)
    if self.start_test_setup_id then
        return {
            result = 'error',
            msg = 'One or more channels on device are being used in a test',
        }
    end
    self.no_test_start = 'Device is performing user requested reboot'
    -- NOTE: We don't unregister as a critical process here in case
    -- a test is actively using channels on this device and some
    -- dufus reboots the device!  Ideally, we should probably just
    -- refuse the request if a test is using this device.
    self.mx:set_led('flashing_red')
    local cmd = string.format(
        "DROPBEAR_PASSWORD=hbmquantumx " ..
        "/usr/bin/dbclient -y hbmuser@%s " ..
        "'sync;sync;/sbin/reboot'", self.host)
    os.execute(cmd)
    ev.Loop.default:unloop()
    return {
        result = 'success',
        msg = "Rebooting",
    }
end

function mxDeviceProxy:change_drd(params, cb)
    if self.start_test_setup_id then
        return {
            result = 'error',
            msg = 'One or more channels on device are being used in a test',
        }
    end
    self.no_test_start = 'Device is performing user DRD change'
    if params[1] then
        local valid_drd = {
            decimal = true,
            binary = true,
            hbm_legacy = true,
        }
        if not valid_drd[params[1]] then
            return {
                result = 'error',
                msg = string.format("Invalid drd: %s", params[1])
            }
        end
        if params[1] == self.mx.control.devinfo.current_drd then
            return {
                result = 'noop',
                msg = string.format("DRD is already set to %s", params[1])
            }
        end
        return self.mx:set_drd(params[1])
    end
    return {
        result = 'error',
        msg = 'Data rate domain not specified',
    }
end

function mxDeviceProxy:rezero_channels(params, cb)
    if params[1] then
        local ret = self.mx:set_channels_zero_offset(params,
            function(r) if cb then cb(r) end end)
        if not ret or not ret.result or not ret.result == 'success' then
            local res = {
                result = 'error',
                msg = 'Failed to set zero offset on one or more channels',
            }
            return res
        end
        return
    end
    local res = {
        result = 'error',
        msg = 'Channel rezeroing information not specified',
    }
    return res
end

function mxDeviceProxy:control(loop, w, http, spec, cb)
    if not spec.request then
        return {
            result = 'error',
            msg = 'Control request with not request specified',
        }
    end
    for _, c in pairs(self.device_channels) do
        if c.connector == spec.connector then
            if not self[spec.request] or
                type(self[spec.request]) ~= 'function' then
                return {
                    result = 'error',
                    msg = string.format("(%s) Invalid request: %s",
                        c.connector, spec.request)
                }
            end
            return self[spec.request](self, spec.params, cb)
        end
    end
    return {
        result = 'error',
        msg = string.format("Connector not found: %s", spec.connector),
    }
end

function mxDeviceProxy:start_channel_stream(w, spec, md_sink, sink)
    -- If a reboot, DRD change, or firmware update has been started, deny test
    -- start
    if self.no_test_start then
        md_sink{error_message = self.no_test_start}
        return
    end
    md_sink{error_message = 'Channel can not be used in a test'}
end

-- Function unsed to send error messages on all channels metadata sinks
function mxDeviceProxy:_send_all_channels_error(msg, start_channel_list, first)
    local start_channel_list = start_channel_list or self.start_channel_list
    for _, c in ipairs(start_channel_list) do
        c.md_sink{error_message = msg}
        if first then break end
    end
    self.start_channel_list = {}
end

-- Function used to send an error message to a channels metadata sink
function mxDeviceProxy:_send_error_on_channel(msg, c)
    c.md_sink{error_message = msg}
    self.start_channel_list = {}
end


-- Function used to clean up when a test is stopped
function mxDeviceProxy:_reset_all_channels()
    if self.config_response_timeout then
        self.config_response_timeout:stop(ev.Loop.default)
    end

    -- Reset sample rates on all channels to 10Hz

    -- Subscribe channels so comet will work
    self.mx.stream:subscribe_all(self.mx.control:get_reference_list())
end

-- Function used to restore channel information after a test start regardless
-- of whether or not the test start was successful or not.
function mxDeviceProxy:_do_channel_cleanup(ref)
    self.mx.stream:detach_sink(ref)
    self.start_test_setup_id = nil
    self.start_ts = nil

    log:log(log.DEBUG, string.format("Restore %s to 10hz", ref))

    local channel_num = ref:lower():match("connector(%d+)")
    self.mx.control:set_channel_sample_rate(channel_num, 10)
    self.mx.control:set_channel_teds_enable(channel_num, config.allow_teds)
end

function mxDeviceProxy:do_post_test_teds()
    local num = 0
    for ref, status in pairs(self.deferred_teds_events) do
        self.teds_events[#self.teds_events + 1] = {ref=ref, status=status}
        num = num + 1
    end
    if num > 0 then
        local sn = sys_notify:new("MX Device Sensor Status' Changed")
        sn:warning("One or more MX device sensors were installed/removed " ..
            "while the test was running.")
    end
    self.deferred_teds_events = {}
-- FIXME:  Enable processing of the deferred TEDS events at some point.
self.teds_events = {}
    self:process_teds_events()
end

-- Function called upon completion of the subscribe request for a channel
function mxDeviceProxy:_test_start_subscribe_handler(signo, ref)
    local info = self.stream_start_info[ref]
    if not info then
        local msg = string.format("Unexpected subscribe callback for %s", ref)
        log:log(log.ERR, msg)
        return
    end

    -- Attach the sink and mark this channel as subscribed
    self.mx.stream:attach_sink(ref, info.w.socket)
    self.stream_start_info[ref].got_subscribe = true

    -- If we haven't gotten subscribe responses for ALL of the channels in the
    -- test, then we are done.
    for _, r in pairs(self.stream_start_info) do
        if not r.got_subscribe then
            return
        end
    end

    -- If test start was aborted while all this channel stuff was occurring,
    -- abort everything, otherwise let it run
    if self.cleanup_was_called or (self.hcr_errors and #self.hcr_errors) then
        local msg = "Test start aborted before test could be started"
        self:_send_all_channels_error(msg, self.saved_start_channel_list, true)
        log:log(log.ERR, msg)
        for ref, info in pairs(self.stream_start_info) do
            self:_do_channel_cleanup(ref)
        end
        self:do_post_test_teds()
    else
        local cu_ref = 0
        for ref, info in pairs(self.stream_start_info) do
            cu_ref = cu_ref + 1
            info.w.cleanup = function()
                    self:_do_channel_cleanup(ref)
                    cu_ref = cu_ref - 1
                    if cu_ref == 0 then
                        self:do_post_test_teds()
                    end
                end
        end
        self.start_test_setup_id = self.start_test_setup_id_pending
        self.start_test_setup_id_pending = nil
    end
end

-- Function used to handle error detected on the data stream while a test is
-- running.
function mxDeviceProxy:_handle_stream_error(msg)
    -- Log the error
    log:log(log.ERR, msg)

    if self.saved_start_channel_list and #self.saved_start_channel_list then
        -- Send error metadata on each channel
        self:_send_error_on_channel(msg, self.saved_start_channel_list[1])

        -- Cleanup all of the channels
        for _, c in ipairs(self.saved_start_channel_list) do
            self:_do_channel_cleanup(c.chan_obj.stream_reference)
        end

        self:do_post_test_teds()
    end

    -- Restore the stream error handler to it's default
    self.mx:restore_stream_error_handler()
end

-- Function called once all channels have been configured and verified.
function mxDeviceProxy:_complete_start_channels()
    log:log(log.DEBUG, "MX Starting channels")
    local start_channel_list = self.saved_start_channel_list

    -- If the debug files stuff is enabled setup the info needed.
    if config.debug_files then
        self.mx.stream.dump_info = {
            file = string.format(
                '/tmp/%s_raw_timestamps.txt', self.name:gsub('-', '_')),
            reference_list = {},
            unit = self.name,
        }
        for r, info in pairs(self.start_test_info) do
            self.mx.stream.dump_info.reference_list[r] = true
        end
    end

    -- Set up stream information for each channel so data can flow at the
    -- proper time
    for r, info in pairs(self.start_test_info) do
        local phase_delay = info.phase_delay
        local st = self.start_ts + self.max_phase_delay
        if not self.start_async then
            self.mx.stream:set_start_time(st, r)
            info.md_sink(info.spec)
            self.mx.stream.subscribe_cb[r] = function(signo, ref)
                    self:_test_start_subscribe_handler(signo, ref)
                end
        else
            local signo = r:match('Connector(%d+)%.')
            signo = ((signo - 1) * 3) + 2
            self.mx.stream:set_start_time(0, r)
            info.md_sink(info.spec)
            self:_test_start_subscribe_handler(signo, r)
        end
    end

    -- Set an error handler to be called if/when any errors are detected on the
    -- data stream.  E.G.  First sample timestamp after test start time, etc.
    self.mx:set_stream_error_handler(function(msg)
            self:_handle_stream_error(msg)
        end)

    local test_id = start_channel_list[1].spec._test_id

    -- If we are NOT in DVM mode, we need to resubscribe to the channels so new
    -- valid first sample timestamps can be obtained for synchronous start
    if not self.start_async then
        -- Get data flowing
        self.mx.stream:subscribe_all(self.mx.control:get_reference_list())
    end
end


-- Function used to handle the response to a configuration request for a channel
function mxDeviceProxy:_handle_config_response(result, c)
    local start_channel_list = self.saved_start_channel_list
    if not result or not result.result or result.result ~= 'success' then
        self.hcr_errors = self.hcr_errors or {}
        self.hcr_errors[#self.hcr_errors + 1] = {
            c = c,
            msg = result.msg
        }
        log:log(log.ERR, result.msg)
    end

    -- Channel was successfully configured, set up the stuff needed to for the
    -- data stream for this channel.
    local phase_delay = self.mx.control.device.phase_delay.fixed or 0
    if c.spec.digital_filter_type ~= 'none' then
        local ft = c.spec.digital_filter_type
        local pd_table = self.mx.control.device.phase_delay[ft]
        if pd_table then
            local freq = tonumber(c.spec.digital_filter_attenuation_frequency)
            if pd_table[freq] then
                phase_delay = phase_delay + pd_table[freq]
            end
        end
    end
    phase_delay = tonumber(phase_delay) / 1000000.0

-- FIXME:  Adjustment of phase delay is disabled
phase_delay = 0
    if phase_delay > self.max_phase_delay then
        self.max_phase_delay = phase_delay
    end

    self.stream_start_info[c.chan_obj.stream_reference] = {
        sink = c.sink,
        phase_delay = phase_delay,
        w = c.w
    }

    self.config_responses = self.config_responses - 1
    if self.config_responses == 0 then
        -- OK we have successful responses to configuration of all of the
        -- channels.
        self.config_response_timeout:stop(ev.Loop.default)
        log:log(log.DEBUG, "MX make sure no channels had setup errors")

        if self.hcr_errors and #self.hcr_errors > 0 then
            for _, v in pairs(self.hcr_errors) do
                v.c.md_sink{error_message = v.msg}
            end
            for _, c in ipairs(self.saved_start_channel_list) do
                self.mx.control:set_channel_sample_rate(c.chan_obj.index, 10)
                self.mx.control:set_channel_teds_enable(c.chan_obj,
                    config.allow_teds)
            end
            self:_reset_all_channels()
            return
        end
        self:_complete_start_channels()
    end
end

function mxDeviceProxy:_setup_channels()
    local start_channel_list = self.saved_start_channel_list
    self.mx.stream:set_start_time(0)

    self.config_responses = #start_channel_list

    -- Set a timeout for configuring the channels
    if not self.config_response_timeout then
        self.config_response_timeout = ev.Timer.new(function(l, t, e)
                local msg = 'Timeout configuring channels'
                log:log(log.ERR, msg)
                self:_send_all_channels_error(msg, start_channel_list, true)
                self:_reset_all_channels()
            end, 30, 30)
    end

    -- Configure each of the channels
    for _, c in ipairs(start_channel_list) do
        local spec = c.spec
        local idx = c.chan_obj.index

        -- Make sure there is no sink attached to this channel
        self.mx.stream:detach_sink(c.chan_obj.stream_reference)

        -- Go configure the channel.  Note, this function returns a synchronous
        -- status indicating if the request was sent off to the MX device.  An
        -- asynchronous response is returned that result in the callback
        -- callback function being called!
        local status =
            self.mx.control:setup_channel(idx, spec, function(result)
                    self:_handle_config_response(result, c)
                end)

        if not status or not status.result or status.result ~= 'success' then
            c.md_sink{error_message = status.msg}
            log:log(log.ERR, status.msg)
            self:_reset_all_channels()
            return
        elseif not self.config_response_timeout:is_active() then
            self.config_response_timeout:start(ev.Loop.default)
            self.config_response_timeout:again(ev.Loop.default)
        end

        -- Disable TEDS on channels used in the test
        self.mx.control:set_channel_teds_enable(idx, false)
    end
end

-- This is the EPOCH on an MX device
local mx_epoch = os.time({
    year = 2000,
    month = 1,
    day = 1,
    hour = 0,
    min = 0,
    sec = 0
})

-- Function used to convert data/time in ISO format to an MX timestamp (NTP)
local function iso_to_qx_ts(iso)
    local time, fraction = { }, nil
    time.year, time.month, time.day, time.hour, time.min, time.sec, fraction =
        iso:match('(%d%d%d%d)%-(%d%d)%-(%d%d)T(%d%d)%:(%d%d)%:(%d%d)%.?(%d*)')
    local seconds = os.time(time)
    local ns = 0
    if fraction then
        local frac_digits = fraction:len()
        ns = fraction * 10^(9 - frac_digits)
    end
    seconds = seconds + (ns / 1000000000)
    return seconds
end


function mxDeviceProxy:_start_channels(start_channel_list)
    -- Clear out any configuration errors
    self.hcr_errors = nil

    -- Restore stream error handler to it's default
    self.mx:restore_stream_error_handler()

    -- Make sure test start has NOT been aborted before we got here
    if #start_channel_list == 0 then
        log:log(log.ERR,
            'Test start aborted before starting channel configuration')
        return
    end

--[[
    -- Make sure we are NOT already running a test
    if self.start_test_setup_id then
        local msg = "Attempt to use device by more than one test at a time"
        self:_send_error_on_channel(msg, start_channel_list[1])
        log:log(log.ERR, msg)
        return
    end
]]
    -- If a reboot, DRD change, or firmware update has been started, deny test
    -- start
    if self.no_test_start then
        local msg = self.no_test_start
        self:_send_error_on_channel(msg, start_channel_list[1])
        log:log(log.ERR, msg)
        return
    end

    -- Now save off the list of channels we are going to be using
    self.saved_start_channel_list = {}
    underscore.extend(self.saved_start_channel_list, start_channel_list)

    -- Make sure the device is synchronized
    if not self.sync_info.synced and not self.disregard_sync_status_changes then
        local msg = "Device not synchronized at test start time"
        self:_send_error_on_channel(msg, start_channel_list[1])
        log:log(log.ERR, msg)
        return
    end

    -- Pending test ID
    self.start_test_setup_id_pending = start_channel_list[1].spec._test_id

    -- Make sure the stream connection for the device is running
    if not self.mx.stream then
        local msg = "Device data stream is not active"
        self:_send_error_on_channel(msg, start_channel_list[1])
        log:log(log.ERR, msg)
        return
    end

    -- Make sure the firmware version on the device is what we expect
    if not config.dont_check_mx_firmware_version and
        self.mx.control.devinfo.fwver ~=
        self.mx.control.devinfo.expected_fwver then
        local msg = string.format(
            "MX device is not running expected firmware version\n" ..
            "Expected version:  %s, actual version:  %s\n",
            self.mx.control.devinfo.expected_fwver,
            self.mx.control.devinfo.fwver)
        self:_send_error_on_channel(msg, start_channel_list[1])
        log:log(log.ERR, msg)
        return
    end

    -- Make sure the DRD for all of the channels in the test are the same
    local drd_wanted
    for _, c in ipairs(start_channel_list) do
        drd_wanted = drd_wanted or c.spec.sample_rate_domain
        if drd_wanted ~= c.spec.sample_rate_domain then
            local msg = string.format(
                "Inconsistant requested sample rate domain on channel %d\n" ..
                "Requested %d, need %d\n", c.index, c.spec.sample_rate_domain,
                drd_wanted)
            self:_send_error_on_channel(msg, start_channel_list[1])
            log:log(log.ERR, msg)
            return
        end
    end

    -- Make sure the DRD matches the DRD of the device
    local current_drd = drd_map[self.mx.control.devinfo.current_drd] or
        'unknown'
    if drd_wanted and drd_wanted ~= current_drd then
        local msg = string.format(
            "Test DRD of '%s' does not match current DRD of '%s'",
            drd_wanted, current_drd)
        self:_send_error_on_channel(msg, start_channel_list[1])
        log:log(log.ERR, msg)
        -- FIXME:  Do we automagically reconfigure the Quantum device for the
        -- desired data rate domain at this point?
        return
    end

    -- Tell stream handling which DRD we are in (for error checking stuff)
    self.mx.stream:data_rate_domain(current_drd)

    log:log(log.DEBUG, "MX Initializing Channels")

    local spec = start_channel_list[1].spec

    -- Set start timestamp
    self.start_ts = iso_to_qx_ts(spec.start_time)

    -- Disable sending data from the stream
    self.mx.stream:set_start_time(0)

    -- Make sure only channels in this test get valid socket FID's
    for k, v in pairs(self.stream_start_info) do
        self.stream_start_info[k].w = {socket = -1}
    end

    self.start_test_info = {}
    self.max_phase_delay = 0

    -- Set up information to be used when test start nears completion
    for _, c in ipairs(start_channel_list) do
        self.start_test_info[c.chan_obj.stream_reference] = {
            spec = c.spec,
            md_sink = c.md_sink,
        }
    end

    -- Unsubscribe from all of the channels, if/when that completes then go
    -- set up the channels
    if not self.start_async then
        -- If not in DVM mode then unsubscribe all of the channels so we can
        -- resubscribe after configuring the channels and get valid first
        -- sample timestamps.
        self.mx.stream:unsubscribe_all(self.mx.control:get_reference_list(),
            function(str)
                local status = self:_setup_channels()
                if not status then
                    -- ERROR!
                end
            end)
        -- Finally, reset the channel list
        self.start_channel_list = {}
    else
        self.start_channel_list = {}
        self:_setup_channels()
    end
end

function mxChannelProxy:initialize(mxDevice, channel_num, is_teds, state)
    self.mxdev = mxDevice
    self.timestamp = 0
    self.index = channel_num
    self.proxy_type = 'mxChannelProxy'
    self.comet_sinks = {}

    if not is_teds or state == 'remove' then
        self.spec = mxDevice.mx.control:get_spec(channel_num)
        self.spec.data_type = 'sequential_float32'
        self.spec.name = string.format('%s_ch_%d', mxDevice.name:gsub('-', '_'),
            channel_num)
        if self.spec.sensor then
            self.spec.sensor.status = 'not connected'
            if self.spec.sensor.use_teds then
                self.spec.sensor.use_teds = tonumber(self.spec.sensor.use_teds)
            end
        else
            self.spec.sensor = {
                status = 'not connected',
                use_teds = 0,
            }
        end
        return
    end

    -- Everything below this point is for TEDS
    self.spec = mxDevice.mx.control:get_teds_spec(self.index)
    self.spec.data_type = 'sequential_float32'
    if not self.spec.sensor.manufacturer or not self.spec.sensor.type or
        not self.spec.sensor.serial_number then
        self.spec = mxDevice.mx.control:get_spec(channel_num)
        self.spec.name = string.format('%s_ch_%d', mxDevice.name:gsub('-', '_'),
            channel_num)
        self.spec.sensor.status = 'connected'
        self.spec.sensor.use_teds = tonumber(self.spec.sensor.use_teds)
        error(string.format(
            "A TEDS device was installed into channel %d and no TEDS data " ..
            "has been supplied by the MX device.  The channel will be " ..
            "treated as though no TEDS device is installed.", channel_num))
        return
    end
    self.spec.name = string.format("%s_%s_%s",
        self.spec.sensor.manufacturer or 'unknown',
        self.spec.sensor.type, self.spec.sensor.serial_number):lower()
    if self.spec.sensor.name then
        self.spec.name = self.spec.name .. "-" .. self.spec.sensor.name
    end
    self.spec.sensor.status = 'connected'
    if self.spec.sensor.use_teds then
        self.spec.sensor.use_teds = tonumber(self.spec.sensor.use_teds)
    end
end

function mxChannelProxy:available_channels(hostname)
    self.stream_reference = self.spec._reference
    self.spec.connector = string.format('@%s', self.proxy_hostname)
    local cnum, base = self.proxy_hostname:match('connector(%d+)%.(mx.+)$')
    self.spec.category = string.format('source/hbm_%s', base:match('(mx.+)-'))
    self.spec.description = string.format("%s Connector %d", base:upper(), cnum)
    self.channel_name = self.spec.connector
    self.spec.valid_controls = {
        initialize_channel = true,
    }
    self.spec.is_logical = false
    local controls = self.mxdev.mx.control.device.controls and
        self.mxdev.mx.control.device.controls.channel or {}
    for c, f in pairs(controls) do
        self.spec.valid_controls[c] = true
        self[c] = f
    end

    if self.spec.sensor and self.spec.sensor.use_teds > 0 and
        self.spec.sensor.status == 'connected' then
        self.spec.description =
            string.format('TEDS %s %s %s device',
            self.spec.sensor.manufacturer, self.spec.sensor.type,
            self.spec.sensor.serial_number)

        if config.use_long_teds_connector_names then
            self.spec.connector =
                string.format('%s-%s-%s@%s', self.spec.sensor.manufacturer,
                    self.spec.sensor.type, self.spec.sensor.serial_number,
                    self.proxy_hostname)
        else
            self.spec.connector = string.format('@%s', self.proxy_hostname)
        end
        self.channel_name = self.spec.connector
        self.spec.electrical_units = nil
    end

    -- Don't include the _reference stuff
    self.spec._reference = nil
    return { self.spec }
end

function mxChannelProxy:start_comet_stream(w, chans, sink)
    if not config.mx_comet_stream_enable then return end
    for _, chan in ipairs(chans) do
        if chan == self.spec.connector then
            self.comet_sinks[w.socket] = sink
            w.cleanup = function()
                self.comet_sinks[w.socket] = nil
            end
        end
    end
end

function mxChannelProxy:initialize_channel(params, cb)
    local spec = params[1]
    local idx = spec.connector:match("connector(%d)")
    self.mxdev.mx.control:setup_channel(idx, spec, function(result)
            cb(result)
        end)
end

function mxChannelProxy:control(loop, w, http, spec, cb)
    if not spec.request then
        return {
            result = 'error',
            msg = 'Control request with not request specified',
        }
    end
    if not self[spec.request] or
        type(self[spec.request]) ~= 'function' then
        return {
            result = 'error',
            msg = string.format("(%s) Invalid request: %s",
                c.connector, spec.request)
        }
    end
    return self[spec.request](self, spec.params, cb)
end

function mxChannelProxy:_abort_test_start(l, t, e)
    self.mxdev.cleanup_was_called = true
    self.start_channel_list = {}
end

-- Function used to start a channel stream for a test
function mxChannelProxy:_start_channel_stream(w, spec, md_sink, sink)
    self.mxdev.cleanup_was_called = nil
    -- First attach a cleanup function to be used until replaced at a later
    -- point in the test start sequence.
    w.cleanup = function()
            self:_abort_test_start()
        end

--[[
    -- For DVM operation, set start_async to true and
    -- disregard_sync_status_changes to true
    --
    -- For asynchronous test purposes, start_async should be true and
    -- disregard_sync_status_changes can be set to either state.
    -- 
]]
    -- A purpose of 'digital_display' means asynchronous start
    self.mxdev.start_async = nil
    self.mxdev.disregard_sync_status_changes = nil
    if spec.purpose and spec.purpose == 'digital_display' then
        self.mxdev.start_async = true
        self.mxdev.disregard_sync_status_changes = true
    end

    -- Set the timestamp for this channel
    self.timestamp = 0

    -- Do some initialization of the list of channels that are in this test
    self.mxdev.start_channel_list = self.mxdev.start_channel_list or {}
    self.mxdev.start_channel_list[spec._test_id] =
        self.mxdev.start_channel_list[spec._test_id] or {}

    -- Make sure all of the "required" parameters have been given
    local spec_required = {
        'source_physical_units',
        'connector',
        'sensor_input'
    }
    for _, k in pairs(spec_required) do
        if not spec[k] then
            local msg = string.format(
                "Required parameter '%s' is not included in setup for %s",
                k, spec.name or 'UNKNOWN NAME')
            log:log(log.ERR, msg)

            -- Send back the message on ALL previously successfully added
            -- channels
            local start_channel_list = self.mxdev.start_channel_list
            for tid, t in pairs(start_channel_list) do
                for _, c in ipairs(t) do
                    c.md_sink{error_message = msg}
                end
            end
            self.mxdev.start_channel_list = {}

            -- Send back for this channel too!
            md_sink{error_message = msg}
            return
        end
    end

    -- Make sure all of the channels starting up are part of the same test id
    local num = 0
    for k, v in pairs(self.mxdev.start_channel_list) do
        num = num + 1
        if num > 1 then
            -- Yikes, more than one test is trying to start up at the same time!
            local msg = 'Can not start more than one test/MX device at a time'
            log:log(log.ERR, msg)

            -- Send back the message on ALL previously successfully added
            -- channels
            local start_channel_list = self.mxdev.start_channel_list
            for tid, t in pairs(start_channel_list) do
                for _, c in ipairs(t) do
                    c.md_sink{error_message = msg}
                end
            end

            -- Send back for this channel too!
            md_sink{error_message = msg}
            return
        end
    end

    local start_channel_list = self.mxdev.start_channel_list[spec._test_id]

    -- Deal wit rationally expressed sample rates
    if spec.sample_rate then
        spec.sample_rate = rational:new(spec.sample_rate):tonumber()
    end

    -- Set up the information that will be used to configure and start this
    -- channel after this function has been called for ALL of this devices
    -- channels in this test!
    start_channel_list[#start_channel_list + 1] = {
        w = w,
        spec = spec,
        md_sink = md_sink,
        sink = sink,
        chan_obj = self,
    }

    -- Make sure changes are saved
    self.mxdev.start_channel_list[spec._test_id] = start_channel_list

    -- If this function has been called for all of the channels on this
    -- device for this test, then we can start the test!
    if #start_channel_list == spec._channel_count then
        self.mxdev:_start_channels(start_channel_list)
    end
end

function mxChannelProxy:start_channel_stream(w, spec, md_sink, sink)
    if self.shunt then
        self:shunt({0},
            function(res)
                if res and res.result and res.result == 'success' then
                    local ok, err = pcall(self._start_channel_stream, self,
                        w, spec, md_sink, sink)
                    if ok then return end
                end
                self.mxdev.start_channel_list = nil
                local msg = "Unable to remove shunt"
                md_sink{error_message = msg}
                log:log(log.ERR, msg)
            end)
        return
    end
    local ok, err = pcall(self._start_channel_stream, self, w, spec, md_sink,
        sink)
    if not ok then
        self.mxdev.start_channel_list = nil
        local msg = string.format("%s: %s", spec.connector, err)
        md_sink{error_message = msg}
        log:log(log.ERR, msg)
        return
    end
end

if arg[1] then
    if mxDeviceProxy:new(arg[1]) then
        ev.Loop.default:loop()
    end
end

log:log(log.INFO, "Proxy exited")
end)

if not _module_ok then
    local log = require 'log'
    log:open('mx-proxy', log.USER)
    log:log(log.ERR, _module_err_msg)
end
