#!/usr/bin/lua

local ffi = require'ffi'
local C = ffi.C
local ev = require'ev'
local ts = require'timestamps'

local config = require 'system_config'

-- Only do this if configured to do so
if not config.timing or not config.timing.ptpv2 or not
    config.timing.ptpv2.pulse_hz then
    os.exit(1)
end

local pulse_hz = tonumber(config.timing.ptpv2.pulse_hz)

local dbus = require'dbus'
local bus = dbus.bus_get('system')
bus:hook_libev_default_loop()

ffi.cdef[[
typedef int     mode_t;
typedef unsigned long   off_t;

int open(const char *pathname, int flags, mode_t mode);
void *mmap(void *addr, size_t length, int prot, int flags, int fd, off_t
    offset);
char *strerror(int errnum);

enum {
    O_RDWR = 2,
    PROT_READ = 1,
    PROT_WRITE = 2,
    MAP_SHARED = 1,
};
]]

local timing = assert(bus:auto_proxy_wait('com.somat.armada.Timing',
                                        '/com/somat/armada/Timing'))

local sys_rate = 40 * 1000000
local fd = C.open('/dev/mem', C.O_RDWR, 0)
assert(fd > 0, string.format("Unable to open /dev/mem: %s",
    ffi.string(C.strerror(ffi.errno()))))
--local map = C.mmap(0x100000, fd, 0xe0000000)
local map = C.mmap(nil, 0x100000, C.PROT_READ + C.PROT_WRITE, C.MAP_SHARED,
    fd, 0xe0000000)
assert(map ~= -1, string.format("Unable to create memory map"))

local function read_32(offset)
    return ffi.cast('unsigned int *', ffi.cast('char *', map) + offset)[0]
end

local function write_32(offset, value)
    ffi.cast('unsigned int *', ffi.cast('char *', map) + offset)[0] = value
end

local st
local function set_last_target()
    st = ts.stamp_to_ts((read_32(0x0061c) * 4294967296) + read_32(0x0062c))
    st.s = st.s + 5
    st.ns = 0
    return ts.ts_to_stamp(st)
end
local last_target = set_last_target()

local inc_value = 1000000000 / (pulse_hz * 2)

timing.TimeJumped:bind(function(xts, xtns, sc, sr, pc, pr)
        last_target = set_last_target()
    end)

ev.Timer.new(function(l, t, e)
        -- Read raw count
        local count = (read_32(0x0061c) * 4294967296) + read_32(0x0062c)
        -- If the current count is greater than the current target then set
        -- a new target.
        if count >= last_target then
            st.ns = st.ns + inc_value
            while st.ns >= 1000000000 do
                st.ns = st.ns - 1000000000
                st.s = st.s + 1
            end
            last_target = ts.ts_to_stamp(st)
            -- Write the new target
            local hi = math.floor(last_target / 4294967296)
            write_32(0x00614, hi)
            write_32(0x00624, math.floor(last_target - (hi * 4294967296)))
        end
    end, 0.001, 1/(pulse_hz*10)):start(ev.Loop.default)

ev.Loop.default:loop()
