local ffi = require 'ffi'

local C = ffi.C

ffi.cdef[[
    struct quorum_latch {
        int required_count;
        int present_count;
        int latched;
        int max_vote_id;
        uint8_t votes[?];
    };
]]

local mt = {}

function mt:initialize(max_vote_id, required_count)
    self.max_vote_id = max_vote_id
    self.required_count = required_count
    self.present_count = 0
    self.latched = 0
    ffi.fill(self.votes, max_vote_id + 1)
    return self
end

function mt:is_latched()
    if self.present_count >= self.required_count then
        self.latched = 1
    end
    return self.latched ~= 0
end

function mt:set_vote(v)
    assert(v >= 0 and v <= self.max_vote_id)
    if self.votes[v] == 0 then
        self.votes[v] = 1
        self.present_count = self.present_count + 1
    end
    return self:is_latched()
end

function mt:get_vote(v)
    assert(v >= 0 and v <= self.max_vote_id)
    return self.votes[v] == 1 and true or false
end

function mt:unset_vote(v)
    assert(v >= 0 and v <= self.max_vote_id)
    if self.votes[v] ~= 0 then
        self.votes[v] = 0
        self.present_count = self.present_count - 1
    end
    return self:is_latched()
end

function mt:__len()
    return self.present_count
end

function mt:__tostring()
    return string.format('[%s %d/%d]', self.latched == 0 and ' ' or '*',
        self.present_count, self.required_count)
end

mt.__index = mt

mt.ctype = ffi.metatype('struct quorum_latch', mt)

local m = {}

function m:new(required_count, max_vote_id)
    max_vote_id = max_vote_id or required_count
    local latch = ffi.new('struct quorum_latch', max_vote_id + 1)
    return latch:initialize(max_vote_id, required_count)
end

return m
