local ffi = require 'ffi'
local C = ffi.C
local stop_tests = require 'tests_killer'
local config = require 'system_config'
local ev = require 'ev'

ffi.cdef[[
    typedef int pid_t;

    void sync(void);
    int kill(pid_t, int);
    int umount2(const char *, int);

    enum {
        SIGUSR1 = 10,
    };

    enum {
        MNT_FORCE = 1,
    };
]]


-- Get the PID for the powerd
local l = io.popen('ps -o pid,comm | grep powerd')
if not l then return end
local powerd_pid
for line in l:lines() do
    powerd_pid = line:match('(%d+) ')
    break
end
l:close()
if not powerd_pid then return end
local ppid = ffi.new('pid_t', tonumber(powerd_pid))

local ledlib = require('led_lib'):new(ev.Loop.default)

-- If rebooting, stop any running tests, when that finishes dismount the
-- storage pack, signal powerd to put LEDs into shutdown pattern, then issue
-- system's reboot.
if arg[0]:find('reboot') then
    if arg[1] and arg[1] == 'error' then
        pcall(ledlib.SetRedLED, ledlib, "on")
    else
        os.execute('touch /var/lib/armada/powered-off;sync')
    end
    pcall(ledlib.SetGreenLED, ledlib, "four_hz")
    stop_tests(false, function()
            C.umount2(config.storage.path, C.MNT_FORCE)
            C.kill(ppid, C.SIGUSR1)
            -- Give the power daemon a bit of time to do it's thing
            ev.Timer.new(function(l, t, e)
                    os.execute('/sbin/reboot')
                end, .5, 0):start(ev.Loop.default)
        end)
    ev.Loop.default:loop()

-- If powering off, signal powerd to perform power down sequence
elseif arg[0]:find('poweroff') then
    C.kill(ppid, C.SIGPWR)
end
