--[[
	Holds some of the sxr import rules, per document SomatXR HBM SDB Sensor Import (Rev 1), 28th December 2014
--]]

--================== Supported Bridge Bridge Enumerations =====

supported_bridges = { 362 , 363 , 364 , 365 , 366 , 367 , 368 , 369 , 371 , 372 , 350 , 351 , 353 , 354 }



--================== Supported Scale Types ======================
scale_types = {
	StrainGage = { 	enums = {0}
					,sxr_name="strain_gage"
					,required = function( cal , l )
									cal.gage_factor = l.GageFactor
									cal.bridge_factor = l.BridgeFactor
									if cal.gage_factor < 0.1 or cal.gage_factor > 10 or cal.bridge_factor == 0 then
										failed = true
									end
								end
				}

	,ZeroSpan = { 	enums = {0,1}
					,sxr_name="defined_zero_span" 
					,required = function( cal , l )
									cal.electrical_zero = l.P1X
									cal.electrical_span = l.P2X
									cal.physical_span = l.P2Y
									if l.P2X == 0 or l.P2Y == 0 then
										failed = true
									end
								end
				}

	,Table2 = { 	enums = {2}
					,sxr_name = "defined_two_point"
					,required = function( cal , l )
									local e1,p1,e2,p2 = string.match( l.ScaleData1 , '~(%d)~(%d)~(%d)~(%d)' )
									cal.electrical = {}
									cal.electrical[1] , cal.electrical[2] = tonumber( e1 ) , tonumber( e2 )
									cal.physical[1] , cal.physical[2] = tonumber( p1 ) , tonumber( p2 )
									if e1 == e2 or p1 == p2 then
										failed = true
									end
								end
				}

	,TwoPoint = { 	enums = {32} 
					,sxr_name="defined_two_point" 
					,required = function( cal , l )
									cal.electrical = {}
									cal.electrical[1] = l.P1X
									cal.electrical[2] = l.P2X
									cal.physical = {}
									cal.physical[1] = l.P1Y
									cal.physical[2] = l.P2Y
									if l.P1X == l.P2X or l.P1Y == l.P2Y then
										failed = true
									end
								end
				}

	,Gain =		{	enums = {64}
					,sxr_name = "defined_slope_intercept"
					,required = function( cal , l )
									cal.offset = tonumber( string.match( l.ScaleData5 , "(.+)~" ) )
									cal.slope = tonumber( string.match( l.ScaleData5 , "~(.+)" ) )
								end
					}

	,Internal =	{	enums = {0,256,512}
					,sxr_name = "internal"
					,required = function( cal , l ) --internal has no requirements, per spec's given by Dennis
									return
								end
					}

}



--==================== Supported Sensor Types =======================================

supported_sensors = {}

	supported_sensors.StrainGage = 	{	name = "StrainGage"
						,categories = { "source/hbm_mx1615b" }
						,elec_units = "mV/V"
						,s_types = {
										scale_types.ZeroSpan
										,scale_types.StrainGage
										,scale_types.TwoPoint
										,scale_types.Table2
										,scale_types.Gain
									}
						,req = function ( s ,  l )
									local supported , failed = false , false
									for k,v in ipairs( supported_bridges ) do
										if v == l.SensorType then
											supported = true
										end
									end
									if l.SensorType == 350 or l.SensorType == 351 then
										if l.SADPara < 0 or l.SADPara > 3 then
											failed = true
										end
									end
									return supported and not failed
								end

						,setup = function( s , l ) --this is really messy, but the rules are kind of all over the place.
													--For some reason, I was told to treat StrainGages and Bridges the same, so who knows.
									local t = l.SensorType
									if supported_sensors.StrainGage.req( s , l ) then
										s.sensor_input = { mode="bridge" , regulation=true , type="full" , wires=6 }
										s.bridge_resistance = 350
										s.bridge_lead_wire_resistance = 0

										if t==365 or t==366 or t==367 or t==354 then
											s.sensor_input.type = 'half'
											s.sensor_input.wires = 5
										elseif t==368 or t==369 or t==371 or t==372 then
											s.sensor_input.type = 'quarter'
											if t==368 or t==369 then
												s.wires = 4
											else s.wires = 3 end
										end

										if t==362 or t==365 or t==368 or t==371 then
											s.bridge_resistance = 120
										elseif t==364 or t==367 then
											s.bridge_resistance = 700
										end

										if t==353 or t==354 or t==350 or t==351 then
											s.bridge_resistance = l.Resistance
											if s.bridge_resistance == 0 then s.bridge_resistance = 350; end
											if t==350 then
												if l.SADPara == 0 then
													s.wires = 6; s.regulation = true
												elseif l.SADPara == 2 then
													s.wires = 4; s.regulation = false
												else
													failed = true
												end
											elseif t==351 then
												s.type = half
												if l.SADPara == 1 then
													s.wires = 5; s.regulation = true;
												elseif l.SADPara == 3 then
													s.wires = 3; s.regulation = false;
												else
													failed = true
												end
											end
										end

										if s.bridge_resistance < 80 or s.bridge_resistance > 1200 then
											failed = true
										end

										if l.ExcitationFrequency == -1 then
											s.excitation_mode = 'dc'
										elseif l.ExcitationFrequency > 0 then
											s.excitation_mode = 'ac_1200_hz_square'
										else
											failed = true
										end

										if l.Excitation >= 5 then s.excitation_range = "5.0v"
										elseif l.Excitation >= 2.5 then s.excitation_range = "2.5v"
										elseif l.Excitation >= 1.0 then s.excitation_range = "1.0v"
										elseif l.Excitation >= 0.5 then s.excitation_range = "0.5v"
										else failed = true end
									end
								end
					}

	supported_sensors.Bridge = supported_sensors.StrainGage;
	supported_sensors.Bridge.name = "Bridge"

	supported_sensors.Voltage = 		{ 	name="Voltage"
						,categories = { "source/hbm_mx1601b" , "source/hbm_mx1615b" }--, "source/hbm_mx1615b" } 
						,elec_units = "V"
						,s_types = {
										scale_types.ZeroSpan
										,scale_types.TwoPoint
										,scale_types.Table2
										,scale_types.Gain
									}
						,req = function( s , l )
									return l.SensorType == 420 and 
										(l.ActiveSupplyVoltage == 0 or (l.ActiveSupplyVoltage <= 30 and l.ActiveSupplyVoltage >= 5) )
								end
						,setup = function( s , l )
									if supported_sensors.Voltage.req( s , l ) then
										s.sensor_input.mode = "volt"
										s.transducer_power = l.ActiveSupplyVoltage
										s.calibration[3] = nil
										local max = 0

										if l.ScaleType == 1 then --ZeroSpan
											max = s.calibration[1].electrical_span
										elseif l.ScaleType == 2 or l.ScaleType == 32 then --Table2/TwoPoint
											max = s.calibration[1].electrical[2]
										elseif l.ScaleType == 64 then --Gain
											max = l.Range
										end

										if max <= 0.1 then
											max = 0.1
										elseif max <= 1 then
											failed = true
										elseif max <= 10 then
											max = 10
										elseif max <= 60 then
											max = 60
										else
											failed = true
										end

										s.sensor_input.max = max
										s.sensor_input.min = max * -1

										if s.transducer_power > 0 or s.sensor_input.max < 6 then
											s._supported_categories = { "source/hbm_mx1601b" }
										end

									else
										failed = true
									end
								end
					}

	supported_sensors.Current = 		{	name="Current"
						,categories = { "source/hbm_mx1601b" } 
						,elec_units = "mA"
						,s_types = { 
										scale_types.ZeroSpan
										,scale_types.TwoPoint 
										,scale_types.Table2 
										,scale_types.Gain 
									}
						,req = function( s , l ) 
										return l.SensorType == 421 and
											( (l.ActiveSupplyVoltage >= 5 and l.ActiveSupplyVoltage <= 30) or l.ActiveSupplyVoltage == 0 )
								end
						,setup = function( s , l )
									if supported_sensors.Current.req( s , l ) then
										s.sensor_input.mode = "4-20mA"
										s.transducer_power = l.ActiveSupplyVoltage
										s.calibration[3] = nil
									else
										failed = true
										print( "Failed check..." )
									end
								end
					}

	supported_sensors.Thermocouple = {	name="Thermocouple"
						,categories = { "source/hbm_mx1609b" } 
						,elec_units = "°C"
						,s_types = {
										scale_types.Internal
									}
						,req = function( s , l )
									return l.SensorType == 451 and ( l.EngUnit == '°C' or l.EngUnit == '°F' )
								end
						,setup = function( s , l )
									if supported_sensors.Thermocouple.req( s , l ) then
										s.sensor_input.mode = "thermocouple"
										s.sensor_input.type = 'k'
									else
										failed = true
									end
								end

					}

	supported_sensors.RTD = 			{	name="RTD"
						,categories = { "source/hbm_mx1615b" } 
						,elec_units = "°C"
						,s_types = {
										scale_types.Internal
									}
						,req = function( s , l )
									return l.SensorType == 501 and l.SADPara == 26 and
										( l.EngUnit == '°C' or l.EngUnit == '°F' )
								end
						,setup = function( s , l )
									if supported_sensors.RTD.req( s , l ) then
										s.sensor_input.mode = "pt100"
									else
										failed = true
									end
								end

					}

	supported_sensors.Resistance = 	{	name="Resistance"
						,categories = { "source/hbm_mx1615b" } 
						,elec_units = "Ohm"
						,s_types = {
										scale_types.ZeroSpan
										,scale_types.TwoPoint
										,scale_types.Table2
										,scale_types.Gain
									}
						,req = function( s , l )
									return l.SensorType == 475 and l.SADPara == 26
								end
						,setup = function( s , l )
									if supported_sensors.Resistance.req( s , l ) then
										s.sensor_input.mode = "resistor"
										s.calibration[3] = nil
									else
										failed = true
									end
								end
					}

	supported_sensors.Potentiometric = 	{	name="Potentiometric"
						,categories = { "source/hbm_mx1615b" }
						,elec_units = "mV/V"
						,s_types = {
										scale_types.ZeroSpan
										,scale_types.TwoPoint
										,scale_types.Table2
										,scale_types.Gain
									}
						,req = function( s , l )
									return l.SensorType == 385 and l.SADPara == 28
								end

						,setup = function( s , l )
									if supported_sensors.Potentiometric.req( s , l ) then
										s.sensor_input.mode = 'potentiometer'
										s.calibration[3] = nil

										if l.ExcitationFrequency == -1 then
											s.excitation_mode = "dc"
										else
											s.excitation_mode = "ac_1200_hz_square"
										end

										if l.Excitation >= 5 then
											s.excitation_range = "5.0v"
										elseif l.Excitation >= 2.5 then
											s.excitation_range = "2.5v"
										elseif l.Excitation >= 1.0 then
											s.excitation_range = "1.0V"
										else failed = true
										end
									else
										failed = true
									end
								end

					}

	supported_sensors.IEPE = 		{	name="IEPE"
						,categories = { "source/hbm_mx1601b" }
						,elec_units = "V"
						,s_types = {
										scale_types.ZeroSpan
										,scale_types.TwoPoint
										,scale_types.Table2
										,scale_types.Gain
									}
						,req = function( s , l )
									return l.SensorType == 581
								end
						,setup = function( s , l )
									if supported_sensors.IEPE.req( s , l ) then
										s.sensor_input.mode = "iepe"
										s.calibration[3] = nil
									else
										failed = true
									end
								end
					}














return supported_sensors