#!/usr/bin/lua

--[[
	"Library" for sddb to sql translation, MDB Translation project
	Started December 9th, 2014, by Eric Ames.
	Refactored and renamed January 8th, 2015

]]

sts = {} --wrapper table





function sts:export_mdb_to_csv( mdb , schema )
	local base = string.match( mdb , "(.+)%." )
	print( base )
	os.execute( 'echo "" > "' .. base .. '.csv"' )

	for table_name , table_info in pairs( schema ) do
		print( table_name )
		local command = [[mdb-export -q '"' -d '|' "]] .. mdb .. '" ' .. table_name .. ' >> "' .. base .. '.csv"'
		os.execute( command )
	end

	local file = io.open( base..".csv" , 'r' )
	local export = file:read("*all")
	file:close()

	return export
end




function sts:parse_csv( text , schema_info )
	local lines , schema = {} , {}

	for table_name , v in pairs( schema_info ) do --get list of expected tables
		schema[ table_name ] = {}
	end

	for newline in string.gmatch( text , "(.-)\n" ) do
		newline = newline .. '|'
		local is_newline_empty = #newline < 10

		if not is_newline_empty then
			local iterated_line , delim_count = parse_line_to_iterated_pair( newline , '|' )
			domovoi:log( newline )

			local is_schema = line_is_schema( iterated_line , schema_info )
			if is_schema then
				schema[ is_schema ] = iterated_line
				for i,v in ipairs( iterated_line ) do
				end
			else
				local current_table , paired_line , current_table_name = nil , {} , nil
				for table_name , table_info in pairs( schema_info ) do
					local diff = table_info.size - delim_count
					if diff < 5 and diff > 0 then
						current_table = schema[ table_name ]
						current_table_name = table_name
						break
					end
				end

				for i , value in ipairs( iterated_line ) do
					value = tonumber(value) or string.match( value , '"(.+)"' ) or value
					paired_line[ string.match( current_table[i] , '"?([%w_]+)"?' ) ] = value
				end

				lines[ paired_line ] = current_table_name
			end
		end
	end
	return lines
end
--===== Helpers ==========
	function parse_line_to_iterated_pair( line , delim )
		local iterated_line , delim_count = {} , 0
		for value in string.gmatch( line , '(.-)%c?'.. delim ) do
			if value == "" then
				value = "null"
			end
			if string.match( value , "[:]" ) and not string.match( value , '"' ) then --special case for dates
				value = '"' .. value .. '"'
			end
			table.insert( iterated_line , tonumber(value) or value )
			delim_count = delim_count + 1
		end
		return iterated_line , delim_count
	end

	function line_is_schema( line , schema_info )
		for i , value in ipairs( line ) do
			for table_name , table_info in pairs( schema_info ) do
				if string.match( value , table_info.sample_column ) then
					return table_name
				end
			end
		end
	end




--[===[
function sts:csv_to_sqlite( text , schema_info , flatfile , source_name )

	local lines = parse_csv( text , schema_info )
	local create_intro , create_chain , insert_intro , insert_chain
	local created_tables = {}

	local chain = ""

	for line , db_table in pairs( lines ) do
		if created_tables[ db_table ] == nil then
			chain = create_table( chain , line , db_table , source_name )
			created_tables[ db_table ] = true
		end

		chain = insert_data( chain , line , db_table , source_name )

	end

	flatfile:write( chain )
	return lines
end
--===== Helpers ============
	function create_table( chain , line , db_table , source_name )
		chain = chain .. "CREATE TABLE IF NOT EXISTS " .. db_table .. " (\n\t"
		chain = chain .. '"Source_File" TEXT \n\t'

		for name , value in pairs( line ) do
			chain = chain .. ',"' .. name .. '"\t' .. check_data_type( value ) .. '\n\t'
		end
		chain = chain .. ');\n\n'
		chain = chain .. "DELETE FROM " .. db_table .. ' WHERE Source_File = "' .. source_name .. '";\n\n'

		return chain
	end

	function insert_data( chain , line , db_table , source_name )
		chain = chain .. "INSERT INTO " .. db_table .. ' VALUES (\n'
		chain = chain .. '"' .. source_name .. '" '

		for name , value in pairs( line ) do
			if tonumber(value) then
				chain = chain .. ', ' .. value
			else
				chain = chain .. ', "' .. value .. '"'
			end
		end

		chain = chain .. ');\n'
		return chain
	end

	function check_data_type( data )
		if tonumber( data ) then
			return 'INTEGER'
		elseif string.match( data , "[:]" ) then
			return 'NUMERIC'
		else
			return 'TEXT'
		end
	end

function sts:insert_into_cx_db( file , db_name )
	local resp = io.popen( 'sqlite3 ' .. db_name .. ' < "' .. file .. '"' ):read("*all")
	if string.len( resp ) > 5 then
		print( resp )
	end
end

--]===]




return sts