#!/usr/bin/lua

--[[
	Second main lua file to translate an HBM Sensor database to a json blob for CX GUI processing.
	Started by Eric Ames on January 6th, 2015

	Status: WIP
	
]]
stj = {} --wrapper table
local command_file_name = "sqlite_export_commands.sql"

local json_f = require "json_format"
local supported_sensors = require "rules"



function stj:translator( all_rows )
	local groups , senpara , blob = {} , {} , {}
	local max_group_index = 0
	failed_count , passed_count = 0 , 0

	--Setup
	for row , db_table in pairs( all_rows ) do --divide the SenPara and Group row entries
		if db_table == "Groups" then
			groups[ row.GroupIndex ] = row
			if row.GroupIndex > max_group_index then max_group_index = row.GroupIndex end
		elseif db_table == "SenPara" then
			table.insert( senpara , row )
		end
	end
	for i = 1, max_group_index do --order the Groups for ipair looping
		if groups[ i ] == nil then
			groups[ i ] = { GroupIndex=i , Name="Missing Group" , ParentIndex = 0 }
		end
	end

	--Actual Translation
	for i , row in ipairs( senpara ) do --finally process every row and translate to blob
		local sxr_row = {}
		local group = groups[ row.GroupIndex ]
		local settings = supported_sensors[ group.Name:match( "%w+" )]
		if not settings then settings = back_check_sensors( group , supported_sensors ) end

		failed = false

		if settings then
			sxr_row = set_fields( row , settings )
		else
			failed = true
		end


		if not failed then
			passed_count = passed_count + 1

		else
			failed_count = failed_count + 1
			sxr_row = {} --reset if failed
			sxr_row.description = row.Description
			sxr_row._supported_categories = {  }
		end

		sxr_row.group_name = group.Name
		table.insert( blob , sxr_row )
	end

	blob = store_rows_by_group( groups , blob ) --orders the translated rows by group hierarchy for GUI
	return blob , failed_count , passed_count , vanilla
end
--===== Helpers ======
	function set_fields( row , settings )
		local sxr_row = json_f:new()
		sxr_row.new = nil

		sxr_row.description 			= row.Description
		sxr_row._supported_categories 	= settings.categories
		sxr_row.electrical_units 		= settings.elec_units
		sxr_row.output_units 			= row.EngUnit
		sxr_row.source_physical_units 	= sxr_row.output_units

		sxr_row.calibration[ 1 ].units = sxr_row.source_physical_units
		if sxr_row.calibration[3] then
			sxr_row.calibration[ 3 ].units = sxr_row.output_units --needs to be removed for some sensor types, RE
		end

		set_sensor_scale( sxr_row , row , settings )
		settings.setup( sxr_row , row )



		return sxr_row
	end

	function set_sensor_scale( sxr_row , lua_row , settings )
		local stype = lua_row.ScaleType
		local cal = sxr_row.calibration[1]
		local scale_to_use


		if stype == 0 and ( settings.name == "StrainGage" or settings.name == "Bridge" ) then
			if lua_row.GageFactor == 0 then
				scale_to_use = settings.s_types[1]
			else
				scale_to_use = settings.s_types[2]
				local w = sxr_row.output_units
				if w ~= 'µm/m' and w ~= 'm/m' and w ~= 'strain' and w~= 'microstrain' then
					failed = true
				else
					sxr_row.source_physical_units = 'µm/m'
				end
			end
		else
			for i , scale_type in ipairs( settings.s_types ) do
				for i , enum in ipairs( scale_type.enums ) do
					if stype == enum then
						scale_to_use = scale_type
					end
				end
			end
		end

		cal.mode = scale_to_use.sxr_name
		scale_to_use.required( cal , lua_row )

	end

	function back_check_sensors( group , supported_sensors )
		for k,v in pairs( supported_sensors ) do
			if group.Name:match( v.name ) then
				return v
			end
		end
	end

	function store_rows_by_group( groups , rows )
		local blob = {}

		for i , group in ipairs( groups ) do
			local has_parent = groups[ group.ParentIndex ]
			local g = {}

			if has_parent then
				set_parent( blob , has_parent , group )
			else
				blob[ group.Name ] = {}
			end
		end

		for i , row in ipairs( rows ) do
			set_row( blob , row )
		end

		return blob
	end

		function set_parent( level , parent , group )
			for name , info in pairs( level ) do --is incomplete ipair, so has gaps BUT is ordered!
				if name == parent.Name then
					level[ name ][ group.Name ] = {}
				elseif type( info ) == 'table' then
					set_parent( level[name] , parent , group )
				end
			end
		end

		function set_row( level , row )
			for name , info in pairs( level ) do
				if name == row.group_name then
					table.insert( level[ name ] , row )
				elseif type( info ) == 'table' then
					set_row( level[name] , row , translated , groups )
				end
			end
		end



function depth_print( target , pad ) --for debugging
	for k,v in pairs( target ) do
		print( pad , k , v )
		if type(v) == 'table' then
			depth_print( v , pad .. "\t" )
		end
	end
end


--Apparently, a similar function already exists in cjson:something or other. Oh well.
function stj:lua_table_to_json( tble )
	local end_f = '\n'
	b = ""--'\b\b\b\b\b\b\b\b'
	local chain = '\n\t{\n\t\t'

	for row , value in pairs( tble ) do
		chain = row_format( chain , row , value , end_f .. '\t' )
	end
	chain = chain:match( '(.+),' ) .. '\n\t}\n'
	return chain
end

--=== Helpers =====
	function row_format( chain , row , value , end_f )
		chain = chain .. '"' .. row .. '": '
		chain = value_format( chain , value , end_f..'\t' )
		return chain
	end

	function value_format( chain , value , end_f )
		local dtype = type( value )

		if dtype == "number" or dtype == "boolean" then
			chain = chain .. tostring(value) .. end_f .. ','
		elseif dtype == "string" then
			chain = chain .. '"' .. value .. '"'..end_f .. ','
		elseif dtype == "table" then
			chain = table_format( chain , value , end_f .. '\t' )
		end
		return chain
	end

	function table_format( chain , value , end_f )
		local is_array = ( #value > 0 )
		local is_kv_array = false

		for k,v in pairs( value ) do
			is_kv_array = true
		end


		if is_array then
			chain = chain .. '[' .. end_f
			for sub_row , sub_value in ipairs( value ) do
				chain = value_format( chain , sub_value , end_f )
			end
			chain = string.match( chain , "(.+),") .. ']' .. end_f .. b .. ','
		elseif is_kv_array then
			chain = chain .. '{'
			for sub_row , sub_value in pairs( value ) do
				chain = row_format( chain , sub_row , sub_value , "" )
			end	
			chain = ( string.match( chain , "(.+),")  ) .. '}' .. end_f .. b .. ','
		else
			chain = chain .. "[]," --empty list/array
		end
		return chain
	end





return stj