-- time_channel.lua - Implements the Time computed channel.

local dbg = require 'dbg'
local ffi = require 'ffi'

local math_min = math.min

local m = {}

function m:run(in_buf, out_buf, sie_writer)
    local doable = 0
    while #in_buf > 0 do
        local out_available = out_buf:available()
        local doable = math_min(1, out_available)
        for i = 0, doable - 1 do
            local sample = in_buf:deq()
            if not sample then break end
            local os = out_buf:enq()
            assert(os, 'buffer overflow in SIE size channel')
            if sie_writer then
                os[0].value[0] = sie_writer:amount_written()
            end
        end
    end
    return doable > 0
end

function m:new()
    self.total_written = 0
    return self
end

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    local attributes = args or {}
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    assert(#self.node.input_edges == 1)
    assert(#self.node.output_edges >= 0
        and #self.node.output_edges <= 1)
    local tv = {data_type = self.node.data_type or
        self.node.meta.data_type,
        category = self.node.meta.category}
    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)'):new()
    ]==], tv)
    return name, value
end

function m:cm_run(name)
    local code = self.compiler:template({output = false, inline = true}, [==[
    |    more = ((=cm_name=):run((=input_edges.1.buf=),
    |+       (=output_edges.1.buf=), self.sw)) or more
    ]==], {cm_name = name}, self.node)
    return code
end

return m
