local cjson = require 'cjson.safe'

local function dbg(fmt, ...)
    io.stderr:write(string.format(fmt, ...))
end

local m = {}

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    return setmetatable(args or {}, cm_mt)
end

function m:cm_initialize(args)
    local compiler = self.compiler
    compiler:oc_declare({op_local = 'comet_values'}, '{}')
    compiler:oc_declare({op_local = 'comet_names'}, '{}')
    local comet_names = {}
    local comet_available = {}
    for idx, edge in ipairs(compiler.edges) do
        if edge.to_node.meta.category == 'sink/comet_stream' then
            comet_names[edge.idx] = edge.from_name
            local ch = {name = edge.from_name}
            for _, tag in ipairs{'description', 'electrical_units',
                    'output_units', 'units'} do
                ch[tag] = edge.from_node:upstream_meta(tag)[1]
            end
            comet_available[edge.idx] = ch
        end
    end

    compiler:push_indent('')
    compiler:oc_declare({op_local = 'comet_available'},
        compiler:render_lua(comet_available))
    compiler:pop_indent()

    local o = {}
    o[#o + 1] = self.compiler:template({output = false}, [==[
    |
    |do
    |    local av = {}
    |    for idx, ch in pairs(comet_available) do
    |        comet_names[idx] = ch.name .. '@' .. self.test_srvname
    |        ch.connector = comet_names[idx]
    |        comet_message[idx] = comet_message[ch.name]
    |        av[#av + 1] = ch
    |    end
    |    self.available_channels = av
    |    self.comet_streams = {}
    |    local comet_streams = self.comet_streams
    |    local function comet_updater(loop, timer, events)
    |        -- Why aren't we using cjson's ability to encode NAN & INF
    |        -- Because it encodes them as numbers rather than as strings!
    |        for k, v in pairs(comet_values) do
    |            if type(v) == 'number' then
    |                if v == 1/0 then comet_values[k] = 'Infinity'
    |                elseif v == -1/0 then comet_values[k] = '-Infinity'
    |                -- WTF?  See weirdness about NAN's and Lua
    |                elseif v ~= v then comet_values[k] = 'NaN'
    |                end
    |            end
    |        end
    |        local v, err = cjson.encode(comet_values)
    |        if v then
    |            v = v .. '\n\x04'
    |            for i = 1, #comet_streams do
    |                comet_streams[i].streambuf:write(v)
    |            end
    |        else
    |            error(err)
    |        end
    |        comet_values = {}
    |    end
    |    self.comet_timer = ev.Timer.new(comet_updater, 0.1, 0.1) -- KLUDGE rate
    |
    |    self:modify_rrs({}, {
    |        {'_armada-connectors', 'PTR', self.test_srvname},
    |        {self.test_srvname, 'TXT', 'available_channels: ' ..
    |            cjson.encode(self.available_channels)},
    |    })
    |end
    ]==])
    return table.concat(o)
end

function m:cm_run(name)
    local o = {}
    for _, edge in ipairs(self.node.input_edges) do
        local tv = {
            edge = edge,
            access = edge.from_node.is_numeric and '.value[0]' or '',
            render = edge.from_node.is_numeric and 'tonumber' or 'tostring',
            clear = '(=edge.buf=):clear()',
            vm = edge.value_malloced,
        }
        o[#o + 1] = self.compiler:template({output = false}, [==[
        |    if #(=edge.buf=) > 0 then
        |        for i = #(=edge.buf=) - 1, 0, -1 do -- Start with newest item
        |            local t = (=edge.buf=):ptr(i)
        |            if not t:is_heartbeat() then
        |                comet_values[comet_names[(=edge.idx=)]] =
        |+                   comet_message[(=edge.idx=)] or
        |+                   (=render=)(t(=access=))
        |                break
        |            end
        |        end
        |!vm     (=edge.buf=):clear()
        |?vm     for i = 1, #(=edge.buf=) do
        |?vm         local t = (=edge.buf=):deq()
        |?vm         C.free(t.value)
        |?vm     end
        |    end
        ]==], tv, self.node)
    end
    return table.concat(o)
end

return m
