local cjson = require 'cjson.safe'

local function dbg(fmt, ...)
    io.stderr:write(string.format(fmt, ...))
end

local m = {}

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    return setmetatable(args or {}, cm_mt)
end

function m:cm_initialize(args)
    local compiler = self.compiler
    compiler:oc_declare({op_local = 'stream_names'}, '{}')
    local stream_names = {}
    local stream_available = {}
    for idx, edge in ipairs(compiler.edges) do
        if edge.to_node.meta.category == 'sink/hbm_stream' then
            stream_names[edge.idx] = edge.from_name
            local ch = {name = edge.from_name}
            for _, tag in ipairs{'description', 'electrical_units',
                    'output_units', 'units'} do
                ch[tag] = edge.from_node:upstream_meta(tag)[1]
            end
            ch.idx = idx
            stream_available[#stream_available + 1] = ch
        end
    end

    compiler:push_indent('')
    compiler:oc_declare({op_local = 'stream_available'},
        compiler:render_lua(stream_available))
    compiler:pop_indent()

    local o = {}
    o[#o + 1] = self.compiler:template({output = false}, [==[
    |
    |do
    |    if self.stream_server then
    |        local i = 1
    |        for idx, ch in pairs(stream_available) do
    |            self.stream_server:add_channel(ch.name, hbm_bufs[i])
    |            i = i + 1
    |        end
    |    end
    |end
    ]==])
    return table.concat(o)
end

function m:cm_run(name)
    local o = {}
    for _, edge in ipairs(self.node.input_edges) do
        local tv = {
            edge = edge,
            access = edge.from_node.is_numeric and '.value[0]' or '',
            render = edge.from_node.is_numeric and 'tonumber' or 'tostring',
            clear = '(=edge.buf=):clear()',
            vm = edge.value_malloced,
            idx = #o + 1
        }
        if edge.from_node.is_message then
            tv.access = ''
        end
        o[#o + 1] = self.compiler:template({output = false}, [==[
        |    while #(=edge.buf=) > 0 do
        |        local v = (=edge.buf=):deq()
        |        if not hbm_bufs[(=idx=)]:enq_value(v[0](=access=)) then
        |            hbm_bufs[(=idx=)]:adjust_attempted(#(=edge.buf=))
        |            break
        |        end
        |    end
        |    (=clear=)
        ]==], tv, self.node)
    end
    return table.concat(o)
end

return m
