local cjson = require 'cjson'

local m = require 'setup_compiler/sink/sie/time_at_level_1d_runtime'

-- setup-compiler module support

local cm_mt = {__index = m}

function m:cm_new(args)
    return setmetatable(args or {}, cm_mt)
end

function m:cm_declare()
    assert(#self.node.input_edges == 1) -- KLUDGE
    local hb = assert(self.node.meta.histogram_bins) -- KLUDGE
    hb = hb[1] and hb[1] or hb -- Allow degenerate {...} or normal [{...}]

    local tv = {}
    tv.bin_spacing = hb.type

    self.node.bin_spacing = hb.type
    self.node.bin_count = hb.number_of_bins

    self.node.meta.bin_type = self.node.meta.bin_type or 'uint32_t'
    local name, value = self.compiler:oc_declare([==[
    |(require 'setup_compiler/(=category=)_runtime'):new('(=bin_type=)')
    ]==], self.node.meta)

    return name, value
end

function m:cm_initialize(args)
    local hb = assert(self.node.meta.histogram_bins) -- KLUDGE
    hb = hb[1] and hb[1] or hb -- Allow degenerate {...} or normal [{...}]

    local code = {}

    local tv = {}

    local input_node = self.compiler:get_node(self.node.meta.input_channels[1])
    local md = input_node.meta
    -- KLUDGE This is all kind of nasty
    tv.input_channel_name = md.name
    tv.input_channel_description = md.description or ''
    tv.sample_rate = md.sample_rate

    code[#code + 1] = self.compiler:template({output = false}, [==[
    |
    |do
    |    local mod = (=mod_name=)
    |    mod:set_aux{compiler = self}
    |    mod.bin_count = (=number_of_bins=) + 2 -- Add underflow, overflow bins
    |    mod.bins = C.calloc(mod.bin_count, ffi.sizeof'(=bin_type=)')
    |    -- There is 1 more bin boundary than bins: -inf, ..., inf
    |    mod.bounds = C.calloc(mod.bin_count + 1, ffi.sizeof'double')
    ]==], hb, self.node, self.node.meta)

    if self.node.bin_spacing == 'evenly_divided' then
        code[#code + 1] = self.compiler:template({output = false}, [==[
        |    mod.lo_bound = (=low_bound=)
        |    mod.hi_bound = (=high_bound=)
        |    mod.bound_span = mod.hi_bound - mod.lo_bound
        |    mod.bound_scale = (mod.bin_count - 2) / mod.bound_span
        |    mod.bounds[0] = -1/0 -- KLUDGE use constant!
        |    mod.bounds[mod.bin_count] = 1/0 -- KLUDGE use constant!
        |    for i = 0, mod.bin_count - 2 do
        |        mod.bounds[i + 1] = mod.lo_bound + i / mod.bound_scale
        |    end
        ]==], hb, self.node, self.node.meta)
    elseif self.node.bin_spacing == 'user_defined' then
        code[#code + 1] = self.compiler:template({output = false}, [==[
        |    local lo = (=high_bound=)
        |    local hi = (=high_bound=)
        |    mod.bounds[0] = -1/0
        |    mod.bounds[(=number_of_bins=) + 2] = 1/0
        |    for i = 1, (=number_of_bins=) + 1 do
        |        mod.bounds[i] =
        |    end
        ]==], hb, self.node, self.node.meta)
    end

    self.node.sie_buf, self.node.sie_buf_idx = self.compiler:oc_sie_buf()

    code[#code + 1] = self.compiler:template({output = false}, [==[
    |    local name = "(=label_name=)" -- KLUDGE
    |    local input_channel_name = "(=input_channel_name=)" -- KLUDGE
    |    local input_channel_description = "(=input_channel_description=)"
    |+       -- KLUDGE
    |    local bounds_bits = 64
    |    local bounds_base = mod:get_histogram_bin_base_channel(bounds_bits)
    |    self.sw:new_buffer(self.sw.scratch_buf_idx, nil,
    |        (mod.bin_count + 1) * bounds_bits / 8,
    |        bounds_bits / 8, 0)
    |    local bbuf = sie_bufs[self.sw.scratch_buf_idx]
    |    bbuf:append(mod.bounds, (mod.bin_count + 1) * bounds_bits / 8)
    |    bbuf:flush(true)
    |    local bounds_ch = mod:histogram_bounds_channel(bounds_base,
    |+       bbuf.group, name)
    |
    |    local bin_bits = 32 -- KLUDGE WRONG BAD
    |    local hist_base = mod:histogram_1d_base_channel{
    |        datamode_name = name,
    |        input_channel_name = input_channel_name,
    |        input_channel_description = input_channel_description,
    |        sample_rate = "(=sample_rate=)",
    |        bin_bits = bin_bits,
    |        index_ch = bounds_ch,
    |        range_min = mod.lo_bound,
    |        range_max = mod.hi_bound,
    |    }
    |    local tv = {}
    |    tv.name = input_channel_name .. "@" .. name
    |    tv.base = hist_base
    |
    |    local dims = 1
    |    self.sw:new_buffer((=sie_buf_idx=), nil,
    |        mod.bin_count * bin_bits / 8,
    |        bin_bits / 8, ffi.sizeof'uint32_t' * (dims + 1))
    |    tv.grp = (=sie_buf=).group
    |    mod:histogram_1d_channel(tv)
    |end
    ]==], hb, tv, self.node, self.node.meta, {sample_rate = 1})

    return table.concat(code)
end

function m:cm_run(name)
    local code = self.compiler:tmpl([==[
    |    (=cm_name=):run_evenly_divided_generic((=input_edges.1.buf=),
    |+       (=input_edges.1.buf=).count)
    ]==], {cm_name = name}, self.node)
    return code
end

function m:cm_finalize(name)
    local code = self.compiler:template({output = false}, [==[
    |do
    |    (=cm_name=):finalize()
    |    local bin_bits = 32 -- KLUDGE WRONG BAD
    |    local hdr = ffi.new('uint32_t[2]')
    |    hdr[0] = 3 -- dim count ???
    |    hdr[1] = (=cm_name=).bin_count
    |    (=sie_buf=):enq(hdr, ffi.sizeof'uint32_t' * 2)
    |    (=sie_buf=):enq((=cm_name=).bins, hdr[1] * bin_bits / 8)
    |    (=sie_buf=):flush()
    |end
    ]==], {cm_name = name}, self.node)
    return code
end

return m
