#!/usr/bin/lua

require 'register_critical'

local root_dir = "/usr/share/web/st"
local lib_dir = "/usr/share/web/st_lib"

local cjson = require "cjson"
local dbg = require "dbg"
local ev = require "ev"
local lfs = require "lfs"
local lpeg = require "lpeg"
local scgi_server = require "scgi_server"
local st_compiler = require "st_compiler"
local u = require "unix"

local ev_loop = ev.Loop.default
local st_compiler = st_compiler:new()

local function read_file(file)
--dbg.debugf("read_file(%s)\n", file)
    local fh = io.open(file)
    assert(fh)
    local got = assert(fh:read("*a"))
    assert(fh:close())
    return got
end

local function read_templates(templates, path, prefix)
    prefix = prefix or ""
    for f in lfs.dir(path) do
--dbg.debugf("file %s\n", f)
        local template_name = string.match(f, "^(.-)%.st$")
        if template_name then
            local fname = string.format("%s/%s", path, f)
            local mode = lfs.attributes(fname, "mode")
            if mode == "file" then
--dbg.debugf("template %s\n", template_name)
                templates[prefix .. template_name] = read_file(fname)
            end
        end
    end
end

local function st_handler(loop, w)
--[==[
    u.write(w.socket,
        "HTTP/1.1 200 OK\n"
        .. "Content-Type: text/plain\n"
        .. "X-Accel-Buffering: no\n"
        .. "\n"
    )
    for k, v in pairs(w.headers) do
        u.write(w.socket, string.format("handler %s => %s\n", k, v))
    end

    for k, v in pairs(w.headers) do
        print(string.format("handler %s => %s", k, v))
    end
--]==]

    local uri = w.headers.DOCUMENT_URI
--print(string.format("uri = %s", uri))

--[[
    local start, req_dir = uri:match("^(/([^/]*/)*[^/]*).html$")
    --if not start or start == "" then
    --    start = uri:match("^/(.-)")
    --end
    if not start or start == "/" then
        --start, req_dir = uri:match("^(/.-)") or "/"
        start, req_dir = uri:match("^(/([^/]*/)*[^/]*).html$")
        if start == "/" or start:match("/$") then
            start = start .. "index"
        end
    end
--]]

--[[
    local P, C = lpeg.P, lpeg.C
    local p_name = P(P(1) - P"/")^0
    local p_name_slash = P(p_name * P"/")
    local p_dir = P(C(P"/" * C(p_name_slash^0) * (p_name - P".html"))
        * (P".html"^0) * P(1)^0)
    local start, req_dir, rest = p_dir:match(uri)

start = start or ""
req_dir = req_dir or ""
rest = rest or ""
print(string.format("start = '%s', req_dir = '%s', rest = '%s'", start, req_dir, rest))
--]]

    local req_dir
    local pos, lpos, start = uri:find("(/[^/]+)%.html$")
    if pos then
        req_dir = uri:sub(1, pos)
    else
        req_dir = uri:match("^/.*/$")
        if not req_dir then
            if uri == "/" then
                req_dir = uri
            else
                req_dir = "" -- KLUDGE to get to 404
            end
        end
        start = req_dir .. "index"
    end
--print(string.format("start='%s', req_dir='%s'", start, req_dir))

--u.write(w.socket, string.format("uri = %s\n", uri))

    local templates = {}
    read_templates(templates, lib_dir, "")
    -- read_templates(templates, root_dir, "/") -- KLUDGE WRONG BAD WRONG
    read_templates(templates, root_dir .. req_dir, "/")
    local compiled, lua_source, tree
        = st_compiler:compile{templates=templates, delimiters={"(=", "=)"}}
    assert(compiled)

--u.write(w.socket, "\n\nlua_source:\n" .. lua_source)

    local model = { -- KLUDGE
        date = os.date("!%a, %d %b %Y %H:%M:%S GMT"),
        request = {},
        content = w.content,
        session_json = cjson.encode(w.session),
        session_key = w.query.session_key,
        session = w.session,
    }
    local request = {}
    for _, k in pairs{'accept_language', 'cookies', 'query', 'headers'} do
        request[k] = w[k]
        model.request[k .. '_json'] = cjson.encode(w[k])
    end
    model.request_json = cjson.encode(request)

    local lua_out = compiled(model, start)
    assert(lua_out)
--print(string.format("\n\nlua_out:\n%s\n", lua_out))
    u.write(w.socket, lua_out)
end

local srv = scgi_server:new{
    ev_loop = ev_loop,
    auto_session = true,
    handler = function(loop, w)
        local ok, err = xpcall(
            function()
                st_handler(loop, w)
            end,
            function(e)
                return e
            end
        )
        if not ok then
            u.write(w.socket,
                "HTTP/1.1 404 Not Found\n"
                .. "Content-Type: text/plain\n"
                .. "\n"
                .. "No such URL: " .. w.headers.DOCUMENT_URI .. "\n"
            )
        end
        w.server.close(loop, w)
    end
}
srv:listen{address="127.0.0.1", port=9002}
srv:loop():loop()
