/*jshint white:false */
(function( $ ) {
    "use strict";
    var undefined;

$.widget( 'ui.somat_bar',
{
    /**
     * @description This attribute defines the natural pixel width of the
     *              gauge. It is used for scale math.
     */
    _normalW     :  200,
    /**
     * @description This attribute defines the natural pixel height of the
     *              gauge. It is used for scale math.
     */
    _normalH     :   60,
    /**
     * @description This attribute represents the natural pixel width of the
     *              gauges display. It is used for percentage math.
     */
    _meterFactor :  178,
    /**
     * @description This attribute represents the natural pixel height of
     *              the gauges bar ticks display when no_svg is TRUE. It is
     *              used for scale math.
     */
    _ticksHeight :   19,

    _baseURL     :  $('script:last').prop('src').replace(/[^\/]*$/, ''),

    /**
     * @description This object stores the instance data of this widget.
     * NOTES:
     *        show_needle : shows a needle instead of a bar when TRUE.
     *                      no_svg must be FALSE.
     */

    //private
    'ready'       :      false,
    'scale_x'     :        1.0,
    'scale_y'     :        1.0,
    'range'       :        100,

    // layout
    'lowerX'      :         10,
    'lowerY'      :         15,
    'upperX'      :        195,
    'upperY'      :         15,
    'unitX'       :        100,
    'unitY'       :         20,
    'legendX'     :        100,
    'legendY'     :         56,
    'rangeFont'   :         40,
    'unitFont'    :         90,
    'legendFont'  :         40,

    // dom
    'svg'         :       null,
    'ticks1'      :       null,
    'ticks2'      :       null,
    'needle'      :       null,
    'bar'         :       null,

    options: {
        'no_svg'            :      false,
        'skin'              :    'white',
        'show_needle'       :       true,
        'top'               :          0,
        'left'              :          0,
        'height'            :         60,
        'width'             :        200,
        'legend'            : 'abcd1234',
        'unit'              :      'psi',
        'range_min'         :          0,
        'range_max'         :        100,
        'reading'           :          0,
        'ticks_major'       :       null,
        'ticks_minor'       :       null,
        'readout_range_min' :       null,
        'readout_range_max' :       null
    },

    /**
     * @description This operation primes the options object for use.
     */
    _prime: function() {
        var options = this.options;
        // update the readouts and range.
        if (options.readout_range_min === null)
            options.readout_range_min = options.range_min;
        if (options.readout_range_max === null)
            options.readout_range_max = options.range_max;

        this.range = options.readout_range_max - options.readout_range_min;

        // normalize the scale of the gauge.
        this.scale_x = options.width / this._normalW;
        this.scale_y = options.height / this._normalH;

        // configure the ticks of the gauge.
        if (options.ticks_major === null)
            options.ticks_major = this.range / 10.0;
        if (options.ticks_minor === null)
            options.ticks_minor = options.ticks_major / 5.0;
    },

    /**
     * @private
     * @description This operation initializes the runtime data and
     *              initiates the asyncronous loading of the gauge GUI.
     */
    _create: function()
    {
        var options = this.options;

        // prime the options passed in.
        this._prime();

        // update the DOM element for this gauge.
        var self = $(this.element[0]);
        self.addClass('dial_gauge').css({
            // top : options.top + 'px;',
            // left: options.left + 'px',
            width: options.width + 'px',
            height: options.height + 'px'
        });
        
        // load in the gauge graphic.
        var handle = this;
        if (options.no_svg)
        {
            $.ajax({url: this._baseURL + 'no_svg/simple_bar_gauge.html',
                    dataType: 'text',
                    success: function(data, textStatus, jqXHR)
                    {
                        handle._ready(data, textStatus);
                    }
            });
        }
        else
        {
            self.svg({loadURL: this._baseURL + 'svg/simple_bar_gauge.svg',
                      onLoad: function(svg, error) {
                              handle._ready(svg, error);} });
        }
    },

    /**
     * @private
     * @description This operation applies any final operations to the DOM
     *              in preparation for runtime use. It expects the return
     *              signature produced by the SVG plug-in callback, or a
     *              success message from the AJAX callback.
     * @param svg This argument provides access to the svg object. or the html
     *            object if no_svg is TRUE.
     * @param msg This argument returns as undefined on success where no_svg
     *            is FALSE. returns as success when no_svg is TRUE. otherwise
     *            it should contain the error message for either mode.
     */
    _ready : function(svg, msg)
    {
        var options = this.options;
        if (typeof(msg) !== 'undefined')
        {
            if (msg === 'success')
            {
                // NOTE: no_svg === TRUE is assumed here.
                var self = $(this.element[0]);
                self.addClass('no_svg_' + options.skin);
                self.html(svg);
                this.reading = $('.gaugeReading', self)[0];
                this.ticks1 = $('.gaugeTicks', self)[0];
                $(this.ticks1).append(
                    "<div class='gaugeBar' style='width:50%'><!--ieb--></div>"
                );
                this.bar = $('.gaugeBar', self)[0];

                this.ready = true;
                this._refresh();
            }
            else
            {
                //if (typeof(console.log) !== 'undefined')
                //{
                //    console.log(msg);
                //}
            }
        }
        else
        {
            var root = svg.root();
            this.svg = svg;
            var dom_channel = $('.gaugeChannel', root)[0];
            this.needle = $('.gaugeNeedle', root)[0];

            // render the bar behind the ticks
            var group;
            if (!options.show_needle)
            {
                group = svg.group(dom_channel,
                                  { 'class' : ('gaugeNeedle0') });
                this.needle = group;
                this.bar = svg.path(group,
                                    'M0,0 L10,10',
                                    { 'class' : 'gaugeLine0',
                                      'fill'  : 'url(#line0_mat)' });
            }
            // add the gauge ticks group
            this.ticks1 = svg.path(dom_channel,
                                   'M0,0 l10,10',
                                   { 'class' : 'gaugeTicks ticks1_' +
                                               options.skin,
                                     'fill'  : 'url(#dark_mat)' });
            this.ticks2 = svg.path(dom_channel,
                                   'M0,0 l10,10',
                                   { 'class' : 'gaugeTicks ticks2_' +
                                               options.skin,
                                     'fill'  : 'url(#dark_mat)' });

            // render the needle in front of the ticks
            if (options.show_needle)
            {
                group = svg.group(dom_channel,
                                  { 'class' : 'gaugeNeedle0' });
                this.needle = group;
                svg.path(group,
                         'M-2.5,17.5 l2,-8 l1,-8 l1,8 l2,8 z',
                         { 'class' : 'needle',
                           'fill'  : 'url(#needle0_mat)' });
            }

            if (options.skin === 'white')
            {
                $('rect.gaugeMask', root).each(function(index)
                    {
                        this.setAttribute('style', 'fill:url(#light_mat)');
                    }
                );
               $('.gaugeFont', root)[0].setAttribute('fill', '#0c0c0d');
            }

            // render the guage ticks
            this._drawTicks();
            this.ready = true;
            this._refresh();
        }
    },

    /**
     * @private
     * @description This operation draws the tick mark elements of the gauge.
     */
    _drawTicks : function()
    {
        var ticks = '';
        var options = this.options;
        var dom_ticks1 = this.ticks1;
        var dom_ticks2 = this.ticks2;
        var _meterFactor = this._meterFactor;
        var x, px;
        if (options.ticks_minor) {
            for (x = options.readout_range_min;
                 x <= options.readout_range_max;
                 x += options.ticks_minor)
            {
                // provide a modulo with a resolution of 0.000
                var modulo = Math.round((x - options.readout_range_min) * 1000.0) %
                             (options.ticks_major * 1000.0);
                if (modulo)
                {
                    px = this._percent(x) * _meterFactor;
                    ticks += 'M' + px + ',0 l0,5 ';
                }
            }
        }
        dom_ticks1.setAttribute('d', ticks);
        if (options.ticks_major)
        {
            for (x = options.readout_range_min;
                 x <= options.readout_range_max;
                 x += options.ticks_major)
            {
                px = (this._percent(x) * _meterFactor);
                ticks += 'M' + px + ',0 l0,10 ';
            }
        }
        dom_ticks2.setAttribute('d', ticks);
    },

    /**
     * @private
     * @description This operation renders the gauge to the GUI in its
     *              current state.
     */
    _present : function()
    {
        var options = this.options;
        var _meterFactor = this._meterFactor;
        var absolute;
        // calculate the percent of the needle.

        if (options.no_svg)
        {
            absolute = this._percent(options.reading) * 100.0;
            this.bar.setAttribute('style',
                                  'width:' + absolute + '%;');
        }
        else
        {
            absolute = _meterFactor * this._percent(options.reading);
            if (options.show_needle)
            {
                this.needle.setAttribute(
                    'transform', 'translate(' + absolute + ')');
            }
            else
            {
                this.bar.setAttribute('d', 'M0.5,1 l0,12 l' +
                                             absolute + ',0 l0,-12 z');
            }
        }
    },

    /**
     * @private
     * @description This operation takes no arguments. It will update all
     *              GUI elements of the SVG Gauge. It will have no effect
     *              if '_ready' is FALSE.
     */
    _refresh : function()
    {
        if (!this.ready)
            return;
        var options = this.options;
        var scale_x = this.scale_x;
        var scale_y = this.scale_y;
        var lowerScale = Math.min(scale_x, scale_y);

        if (options.no_svg)
        {
            var root = $(this.element[0]);

            $('.gaugeUnit', root).text(options.unit).attr(
                'style', 'font-size:' +
                         (this.unitFont * lowerScale) + '%;');
            $('.gaugeLegend', root).text(options.legend).attr(
                'style', 'font-size:' +
                         (this.legendFont * lowerScale) + '%;');
            $('.gaugeTicks', root).attr(
                'style', 'height:' +
                         (this._ticksHeight * scale_y) + 'px;');
            $('.gaugeLower', root).text(options.readout_range_min);
            $('.gaugeUpper', root).text(options.readout_range_max);
        }
        else
        {
            var svg_root = this.svg.root();
            // update the rendering of the gauge.
            $('.gauge', svg_root).attr(
                'transform', 'scale(' + scale_x + ',' + scale_y + ')'
            );

            $('.gaugeLower', svg_root).text(options.readout_range_min).attr(
                {'x'         : this.lowerX * scale_x,
                 'y'         : this.lowerY * scale_y,
                 'font-size' : (this.rangeFont * lowerScale) + '%'});

            $('.gaugeUpper', svg_root).text(options.readout_range_max).attr(
                {'x'         : this.upperX * scale_x,
                 'y'         : this.upperY * scale_y,
                 'font-size' : (this.rangeFont * lowerScale) + '%'});

            $('.gaugeUnit', svg_root).text(options.unit).attr(
                {'x'         : this.unitX * scale_x,
                 'y'         : this.unitY * scale_y,
                 'font-size' : (this.unitFont * lowerScale) + '%'});

             $('.gaugeLegend', svg_root).text(options.legend).attr(
                {'x'         : this.legendX * scale_x,
                 'y'         : this.legendY * scale_y,
                 'font-size' : (this.legendFont * lowerScale) + '%'});
        }
        this.set_reading(options.reading);
    },

    /**
     * @private
     * @description This operation calculates the placement of a reading
     *              as a percentage of the range.
     * @param reading This argument defines the reading to be evaluated.
     * @return This operation returns a value in % where 'min' = '0'
     *         percent and 'max' = '100' percent.
     */
    _percent: function(reading)
    {
        var options = this.options;
        return ((reading - options.readout_range_min) / this.range);
    },

    /**
     * @description This operation updates the current configuration of
     *              the gauge and refreshes the display.
     * @param new_config This argument expects and object containing key
     *                   value pairs to be updated. Unspecified
     *                   configuration items will not be changed.
     */
    set_configuration : function( new_config )
    {
        if ( new_config ) {
            var options = this.options;
            // Merge in the new configuration items.
            $.extend( options, new_config );
            // prime the options passed in.
            this._prime();

            // skip past render updates if not ready.
            if (this.ready && !options.no_svg)
            {
                // update the root to the new size.
                $('.gauge', this.svg.root()).parent().attr({
                    height: options.height,
                    width: options.width
                });
                this._drawTicks();
            }

            // redraw the gauge.
            this._refresh();
        }
    },

    /**
     * @description This operation updates the current reading of the
     *              gauge and updates the GUI when '_ready' is TRUE.
     * @param updated_reading This argument defines the new FLOAT value
     *                        that this gauge should reflect.
     */
    set_reading : function(updated_reading)
    {
        var options = this.options;
        var updated = parseFloat(updated_reading);
        // peg the needle.
        if (updated < options.readout_range_min)
            updated = options.readout_range_min;
        else if (updated > options.readout_range_max)
            updated = options.readout_range_max;
        // apply the value.
        options.reading = updated;
        if (this.ready)
        {
            this._present();
        }
    },

    /**
     * @private
     * @description This operation finalizes the object for destruction.
     */
    destroy: function() {
        var self = $(this.element[0]);
        self.svg('destroy');

        $.Widget.prototype.destroy.apply(this, arguments);
    }
});

}( jQuery ));
