(function($) {
    "use strict";
    /*global alert, Slick, uuid */

    /***
     * A sample Model implementation.
     * Provides a filtered view of the underlying data.
     *
     * Relies on the data item having an "id" property uniquely identifying it.
     */
    function DataView(options) {
        var self = this;

        var defaults = {
        };


        // private
        var idProperty = "id";  // property holding a unique row id
        var items = [];         // data by index
        var rows = [];          // data by row
        var idxById = {};       // indexes by id
        var rowsById = null;    // rows by id; lazy-calculated
        var filter = null;      // filter function
        var updated = null;     // updated item ids
        var suspend = false;    // suspends the recalculation
        var sortAsc = true;
        var fastSortField;
        var sortComparer;
        var refreshHints = {};
        var prevRefreshHints = {};
        var filterArgs;
        var filteredItems = [];
        var compiledFilter;
        var compiledFilterWithCaching;
        var filterCache = [];

        // tracked_db back end
        var backend = null;
        var db_url = null;

        var pagesize = 0;
        var pagenum = 0;
        var totalRows = 0;

        // events
        var onRowCountChanged = new Slick.Event();
        var onRowsChanged = new Slick.Event();
        var onDelta = new Slick.Event();
        var onPagingInfoChanged = new Slick.Event();
        var onBackendChanged = new Slick.Event();

        options = $.extend(true, {}, defaults, options);


        function beginUpdate() {
            suspend = true;
        }

        function endUpdate() {
            suspend = false;
            refresh();
        }

        function setRefreshHints(hints){
            refreshHints = hints;
        }

        function setFilterArgs(args) {
            filterArgs = args;
        }

        function updateIdxById(startingIndex) {
            startingIndex = startingIndex || 0;
            var id;
            for (var i = startingIndex, l = items.length; i < l; i++) {
                id = items[i][idProperty];
                if (id == null) {
                    throw "Each data element must implement a unique 'id' property";
                }
                idxById[id] = i;
            }
        }

        function ensureIdUniqueness() {
            var id;
            for (var i = 0, l = items.length; i < l; i++) {
                id = items[i][idProperty];
                if (id == null || idxById[id] !== i) {
                    throw "Each data element must implement a unique 'id' property";
                }
            }
        }

        function getItems() {
            return items;
        }

        function setItems(data, objectIdProperty) {
            if (objectIdProperty != null) {
                idProperty = objectIdProperty;
            }
            items = filteredItems = data;
            idxById = {};
            updateIdxById();
            ensureIdUniqueness();
            refresh();
        }

        function setBackend(db) {
            // KLUDGE assert db.url
            backend = db;
            if (backend.db_id != null) {
                db_url = backend.url + backend.db_id + '/';
                onBackendChanged.notify(backend);
                loadFromBackend();
            }
        }

        function createBackendSetup(setup_id, ajax_args) {
            // KLUDGE do something here
            if (ajax_args == null)
                ajax_args = {};
            backendRequest($.extend({
                type: 'PUT',
                url: backend.url + setup_id
            }, ajax_args));
        }

        function receiveBackendItems(data) {
            updated = {};
            for (var i = 0, l = data.length; i < l; i++) {
                var item = data[i];
                // KLUDGE do filtering stuff here
                updated[item[idProperty]] = true;
            }
            setItems(data);
        }

        function loadFromBackend() {
            if (backend == null)
                return;
            // KLUDGE get state, refresh(), etc?
            backendRequest({
                type: 'GET',
                url: db_url + '_all_docs',
                success: function (data, status, xhr) {
                    if (data && data.error == null) {
                        receiveBackendItems(data);
                    } else {
                        alert("loadFromBackend() ERROR PLACEHOLDER, db_id=" +
                            backend.db_id);
                        // KLUDGE Maybe do something else here...
                    }
                }
            });
        }

        function setPagingOptions(args) {
            if (args.pageSize != null) {
                pagesize = args.pageSize;
                pagenum = Math.min(pagenum, Math.ceil(totalRows / pagesize));
            }

            if (args.pageNum != null) {
                pagenum = Math.min(args.pageNum, Math.ceil(totalRows / pagesize));
            }

            onPagingInfoChanged.notify(getPagingInfo(), null, self);

            refresh();
        }

        function getPagingInfo() {
            return {pageSize: pagesize, pageNum: pagenum, totalRows: totalRows};
        }

        function sort(comparer, ascending) {
            sortAsc = ascending;
            sortComparer = comparer;
            fastSortField = null;
            if (ascending === false) {
                items.reverse();
            }
            items.sort(comparer);
            if (ascending === false) {
                items.reverse();
            }
            idxById = {};
            updateIdxById();
            refresh();
        }

        /***
        * Provides a workaround for the extremely slow sorting in IE.
        * Does a [lexicographic] sort on a give column by temporarily overriding Object.prototype.toString
        * to return the value of that field and then doing a native Array.sort().
        */
        function fastSort(field, ascending) {
            sortAsc = ascending;
            fastSortField = field;
            sortComparer = null;
            var oldToString = Object.prototype.toString;
            Object.prototype.toString = (typeof field === "function") ? field :
                function () { return this[field]; };
            // an extra reversal for descending sort keeps the sort stable
            // (assuming a stable native sort implementation, which isn't true in some cases)
            if (ascending === false) {
                items.reverse();
            }
            items.sort();
            Object.prototype.toString = oldToString;
            if (ascending === false) {
                items.reverse();
            }
            idxById = {};
            updateIdxById();
            refresh();
        }

        function reSort() {
            if (sortComparer) {
                sort(sortComparer, sortAsc);
            } else if (fastSortField) {
                fastSort(fastSortField, sortAsc);
            }
        }

        function setFilter(filterFn) {
            filter = filterFn;
            compiledFilter = compileFilter();
            compiledFilterWithCaching = compileFilterWithCaching();
            refresh();
        }

        function getItemByIdx(i) {
            return items[i];
        }

        function getIdxById(id) {
            return idxById[id];
        }

        function ensureRowsByIdCache() {
            if (!rowsById) {
                rowsById = {};
                for (var i = 0, l = rows.length; i < l; i++) {
                    rowsById[rows[i][idProperty]] = i;
                }
            }
        }

        function getRowById(id) {
            ensureRowsByIdCache();
            return rowsById[id];
        }

        function getItemById(id) {
            return items[idxById[id]];
        }

        function backendRequest(req) {
            var nr = $.extend({
                error: backendRequestError
            }, req);
            $.command('#ajax', nr);
        }

        function backendRequestError(data, status, xhr) {
            alert("BACKEND FAILURE PLACEHOLDER");
        }

        function createDocument(meta, setup_id) {
            if (setup_id == null)
                setup_id = uuid();
            backend.db_id = setup_id;
            createBackendSetup(backend.db_id);
            setBackend(backend);
        }

        // action = { method:, description:, item:}
        // TODO: Queue up requests, retire from queue on OK, replay on error
        function updateBackend(action, meta) {
            if (backend == null)
                return;
            if (backend.db_id == null) {
                createDocument(meta);
            }
            if (meta && meta.origin === 'remote')
                return;
            var description = action.description;
            if (meta && meta.description)
                description = meta.description;
            var json_data = $.toJSON(action.item);
            backendRequest({
                type: action.method,
                url: db_url + action.item[idProperty] +
                    '?description=' + encodeURI($.toJSON(description)),
                cache: true, // See http://bugs.jquery.com/ticket/4898
                contentType: 'application/json',
                processData: false,
                data: json_data,
                dataType: 'json'
            });
        }

        function baseDelta(meta) {
            var delta = (meta == null) ? {} : _.clone(meta);
            delta.old = {};
            delta.new = {};
            return delta;
        }

        function updateItem(id, item, meta) {
            if (idxById[id] == null || id !== item[idProperty])
                throw "Invalid or non-matching id";

            var current = items[idxById[id]];
            var delta = baseDelta(meta);
            delta.old[item.id] = current;
            delta.new[item.id] = item;
            delta.description = delta.description ||
                ["Update {1}/{2}", backend ? backend.db_id : null, item.id];
            deltaItems(delta);
        }

        function insertItem(insertBefore, item, meta) {
            var delta = baseDelta(meta);
            delta.old[item.id] = null;
            delta.new[item.id] = item;
            delta.positions = {};
            delta.positions[item.id] = insertBefore;
            delta.description = delta.description ||
                ["Insert {1}/{2} at {3}", backend ? backend.db_id : null,
                    item.id, insertBefore];
            deltaItems(delta);
        }

        function addItem(item, meta) {
            var delta = baseDelta(meta);
            delta.old[item.id] = null;
            delta.new[item.id] = item;
            delta.description = delta.description ||
                ["Add {1}/{2}", backend ? backend.db_id : null, item.id];
            deltaItems(delta);
        }

        function deleteItem(id, meta) {
            var idx = idxById[id];
            if (idx == null) {
                throw "Invalid id";
            }

            var delta = baseDelta(meta);
            delta.old[id] = items[idx];
            delta.new[id] = null;
            delta.description = delta.description ||
                ["Delete {1}/{2}", backend ? backend.db_id : null, id];
            deltaItems(delta);
        }

        function deltaItems(delta) {
            var d_old = delta.old;
            var d_new = delta.new;
            var out_old = {};
            var out_new = {};
            var err = "";
            var changed = false;

            // For local edits so the caller doesn't have to laboriously
            // construct a delta for no reason
            if (d_old == null && delta.origin !== "remote") {
                d_old = {};
                _.each(d_new, function(to, id) {
                    var cur = items[idxById[id]];
                    if (cur != null) {
                        d_old[id] = {};
                        _.each(to, function(v, k) {
                            d_old[id][k] = cur[k];
                        });
                    } else {
                        d_old[id] = null;
                    }
                });
            }

            if (!_.isEmpty(_.difference(_.keys(d_old), _.keys(d_new)))) {
                err += "Broken input to deltaItems\n";
            }

            // KLUDGE delta.remote take precedence on conflict?
            // Highlight on conflict?

            _.each(d_old, function(old, id) {
                // delta.cur is a way of getting around SlickGrid's
                // having mutated our item already so that we can't
                // do the delta. Outside, we'll be squirreling away
                // the item in onBeforeEditCell, and then setting
                // cur from that in onCellChange when deltaItems() is called
                var cur = (delta.cur != null) ?
                    delta.cur[id] : items[idxById[id]];
                var to = d_new[id];
                var idx;

                // Short-circuit any edits to the current state
                if (_.isEqual(cur, to))
                    return;

                if (old == null) {
                    if (cur == null) {
                        out_old[id] = null;
                        out_new[id] = to;
                        changed = true;
                        var insertBefore = items.length;
                        if (delta.positions) {
                            idx = delta.positions[id];
                            if (typeof idx === "number" &&
                                    idx >= 0 && idx <= insertBefore)
                                insertBefore = idx;
                        }
                        items.splice(insertBefore, 0, to);
                        updateIdxById(insertBefore);
                    } else {
                        err += "Conflict at " + id +
                            ", insert of already existing item\n";
                    }
                } else if (to == null) {
                    if (cur != null) {
                        out_old[id] = old;
                        out_new[id] = null;
                        changed = true;
                        idx = idxById[id];
                        if (idx == null) {
                            throw "Invalid id";
                        }
                        delete idxById[id];
                        items.splice(idx, 1);
                        updateIdxById(idx);
                    } else {
                        err += "Conflict at " + id +
                            ", delete of nonexisting item\n";
                    }
                } else {
                    var new_item;
                    _.each(old, function(v, k) {
                        // Only do something if what's requested isn't what is
                        if (to[k] !== cur[k]) {
                            if (cur[k] === v) {
                                if (!out_old[id]) {
                                    out_old[id] = {};
                                    out_new[id] = {};
                                    new_item = _.clone(cur);
                                }
                                out_old[id][k] = v;
                                out_new[id][k] = to[k];
                                new_item[k] = to[k];
                            } else {
                                err += "Conflict at " + id + "/" + k +
                                    " [" + cur[k] + " <=> " + v + "]\n";
                            }
                        }
                    });
                    if (new_item) {
                        changed = true;
                        if (idxById[id] == null || id !== new_item[idProperty])
                            throw "Invalid or non-matching id";

                        items[idxById[id]] = new_item;
                        if (!updated) {
                            updated = {};
                        }
                        updated[id] = true;
                    }
                }
            });

            if (err !== "") {
            }

            if (!changed)
                return;

            refresh();

            // KLUDGE construct description above for auto-coalesce use:
            // (as JSON strings)
            // ["INSERT {1}", ITEM-ID] // Display/sort position???
            // ["Add row {1}", ITEM-ID] // Display/sort position???
            // ["Add channel {1}", ITEM-ID] // Display/sort position???
            // ["DELETE {1}", ITEM-ID]
            // ["Remove row {1}", ITEM-ID]
            // ["Remove channel {1}", ITEM-ID]
            // ["Edit row {1} field {2}", ITEM-ID, column-name]
            // delta coalesce argument - false, "auto"
            // auto only coalesces when descriptions are identical and
            // deltas are compatible - auto probably default

            var out_delta = $.extend({}, delta, {old: out_old, new: out_new});
            if (backend != null && delta.origin !== 'remote') {
                if (backend.db_id == null) {
                    createDocument(delta); // KLUDGE???
                }
                var json_data = $.toJSON(out_delta);
                backendRequest({
                    type: 'POST',
                    url: db_url + '_delta' +
                        '?description=' + // KILL
                        encodeURI($.toJSON(delta.description || "deltaItems")),
                    cache: true, // See http://bugs.jquery.com/ticket/4898
                    contentType: 'application/json',
                    processData: false,
                    data: json_data,
                    dataType: 'json'
                });
            }
            onDelta.notify(out_delta);
            return out_delta;
        }

        function getLength() {
            return rows.length;
        }

        function getItem(i) {
            return rows[i];
        }

        function getItemMetadata(i) {
            var item = rows[i];
            if (item == null) {
                return null;
            }
            return null;
        }

        function getFunctionInfo(fn) {
            var fnRegex = /^function[^(]*\(([^)]*)\)\s*\{([\s\S]*)\}$/;
            var matches = fn.toString().match(fnRegex);
            return {
                params: matches[1].split(","),
                body: matches[2]
            };
        }
        
        function compileFilter() {
            /*jshint evil: true */
            var filterInfo = getFunctionInfo(filter);

            var filterBody = filterInfo.body
                .replace(/return false[;}]/gi, "{ continue _coreloop; }")
                .replace(/return true[;}]/gi, "{ _retval[_idx++] = $item$; continue _coreloop; }")
                .replace(/return ([^;}]+?);/gi,
                    "{ if ($1) { _retval[_idx++] = $item$; }; continue _coreloop; }");

            // This preserves the function template code after JS compression,
            // so that replace() commands still work as expected.
            var tpl = [
                //"function(_items, _args) { ",
                "var _retval = [], _idx = 0; ",
                "var $item$, $args$ = _args; ",
                "_coreloop: ",
                "for (var _i = 0, _il = _items.length; _i < _il; _i++) { ",
                "$item$ = _items[_i]; ",
                "$filter$; ",
                "} ",
                "return _retval; "
                //"}"
            ].join("");
            tpl = tpl.replace(/\$filter\$/gi, filterBody);
            tpl = tpl.replace(/\$item\$/gi, filterInfo.params[0]);
            tpl = tpl.replace(/\$args\$/gi, filterInfo.params[1]);

            var fn = new Function("_items,_args", tpl);
            fn.displayName = "compiledFilter";
            return fn;
        }

        function compileFilterWithCaching() {
            var filterInfo = getFunctionInfo(filter);
            
            var filterBody = filterInfo.body
                // {{{ %-match
                .replace(/return false[;}]/gi, "{ continue _coreloop; }")
                .replace(/return true[;}]/gi, "{ _cache[_i] = true;_retval[_idx++] = $item$; continue _coreloop; }")
                .replace(/return ([^;}]+?);/gi,
                    "{ if ((_cache[_i] = $1)) { _retval[_idx++] = $item$; }; continue _coreloop; }");

            // This preserves the function template code after JS compression,
            // so that replace() commands still work as expected.
            var tpl = [
                //"function(_items, _args, _cache) { ",
                "var _retval = [], _idx = 0; ",
                "var $item$, $args$ = _args; ",
                "_coreloop: ",
                "for (var _i = 0, _il = _items.length; _i < _il; _i++) { ",
                "$item$ = _items[_i]; ",
                "if (_cache[_i]) { ",
                "_retval[_idx++] = $item$; ",
                "continue _coreloop; ",
                "} ",
                "$filter$; ",
                "} ",
                "return _retval; "
                //"}"
            ].join("");
            tpl = tpl.replace(/\$filter\$/gi, filterBody);
            tpl = tpl.replace(/\$item\$/gi, filterInfo.params[0]);
            tpl = tpl.replace(/\$args\$/gi, filterInfo.params[1]);

            /*jshint evil:true */
            var fn = new Function("_items,_args,_cache", tpl);
            fn.displayName = "compiledFilterWithCaching";
            return fn;
        }

        function getFilteredAndPagedItems(items) {
            if (filter) {
                if (refreshHints.isFilterNarrowing) {
                    filteredItems = compiledFilter(filteredItems, filterArgs);
                } else if (refreshHints.isFilterExpanding) {
                    filteredItems = compiledFilterWithCaching(items, filterArgs, filterCache);
                } else if (!refreshHints.isFilterUnchanged) {
                    filteredItems = compiledFilter(items, filterArgs);
                }
            } else {
                // special case:  if not filtering and not paging, the resulting
                // rows collection needs to be a copy so that changes due to sort
                // can be caught
                filteredItems = pagesize ? items : items.concat();
            }

            // get the current page
            var paged;
            if (pagesize) {
                if (filteredItems.length < pagenum * pagesize) {
                    pagenum = Math.floor(filteredItems.length / pagesize);
                }
                paged = filteredItems.slice(pagesize * pagenum, pagesize * pagenum + pagesize);
            } else {
                paged = filteredItems;
            }

            return {totalRows: filteredItems.length, rows: paged};
        }

        function getRowDiffs(rows, newRows) {
            var item, r, eitherIsNonData, diff = [];
            var from = 0, to = newRows.length;

            if (refreshHints && refreshHints.ignoreDiffsBefore) {
                from = Math.max(0,
                    Math.min(newRows.length, refreshHints.ignoreDiffsBefore));
            }

            if (refreshHints && refreshHints.ignoreDiffsAfter) {
                to = Math.min(newRows.length,
                    Math.max(0, refreshHints.ignoreDiffsAfter));
            }

            for (var i = from, rl = rows.length; i < to; i++) {
                if (i >= rl) {
                    diff[diff.length] = i;
                } else {
                    item = newRows[i];
                    r = rows[i];

                    if ((eitherIsNonData = (item.__nonDataRow) || (r.__nonDataRow)) &&
                            item.__updated ||
                            item[idProperty] !== r[idProperty] ||
                            (updated && updated[item[idProperty]])
                        ) {
                        diff[diff.length] = i;
                    }
                }
            }
            return diff;
        }

        function recalc(_items) {
            rowsById = null;

            if (refreshHints.isFilterNarrowing !== prevRefreshHints.isFilterNarrowing ||
                refreshHints.isFilterExpanding !== prevRefreshHints.isFilterExpanding) {
                filterCache = [];
            }

            var filteredItems = getFilteredAndPagedItems(_items);
            totalRows = filteredItems.totalRows;
            var newRows = filteredItems.rows;

            var diff = getRowDiffs(rows, newRows);

            rows = newRows;

            return diff;
        }

        function refresh() {
            if (suspend) {
                return;
            }

            var countBefore = rows.length;
            var totalRowsBefore = totalRows;

            var diff = recalc(items, filter); // pass as direct refs to avoid closure perf hit

            // if the current page is no longer valid, go to last page and recalc
            // we suffer a performance penalty here, but the main loop (recalc) remains highly optimized
            if (pagesize && totalRows < pagenum * pagesize) {
                pagenum = Math.floor(totalRows / pagesize);
                diff = recalc(items, filter);
            }

            updated = null;
            prevRefreshHints = refreshHints;
            refreshHints = {};

            if (totalRowsBefore !== totalRows) {
                onPagingInfoChanged.notify(getPagingInfo(), null, self);
            }
            if (countBefore !== rows.length) {
                onRowCountChanged.notify({previous: countBefore, current: rows.length}, null, self);
            }
            if (diff.length > 0) {
                onRowsChanged.notify({rows: diff}, null, self);
            }
        }


        return {
            // methods
            "setBackend":       setBackend,
            "backendRequest":   backendRequest,
            "createBackendSetup":  createBackendSetup,
            "loadFromBackend":  loadFromBackend,
            "backendRequestError": backendRequestError,
            "beginUpdate":      beginUpdate,
            "endUpdate":        endUpdate,
            "setPagingOptions": setPagingOptions,
            "getPagingInfo":    getPagingInfo,
            "getItems":         getItems,
            "setItems":         setItems,
            "setFilter":        setFilter,
            "sort":             sort,
            "fastSort":         fastSort,
            "reSort":           reSort,
            "getIdxById":       getIdxById,
            "getRowById":       getRowById,
            "getItemById":      getItemById,
            "getItemByIdx":     getItemByIdx,
            "setRefreshHints":  setRefreshHints,
            "setFilterArgs":    setFilterArgs,
            "refresh":          refresh,
            "createDocument":   createDocument,
            "updateItem":       updateItem,
            "insertItem":       insertItem,
            "addItem":          addItem,
            "deleteItem":       deleteItem,
            "deltaItems":       deltaItems,

            // data provider methods
            "getLength":        getLength,
            "getItem":          getItem,
            "getItemMetadata":  getItemMetadata,

            // events
            "onRowCountChanged":    onRowCountChanged,
            "onRowsChanged":        onRowsChanged,
            "onDelta":              onDelta,
            "onPagingInfoChanged":  onPagingInfoChanged,
            "onBackendChanged":     onBackendChanged
        };
    }

    $.extend(true, window, {
        Slick: {
            Data: {
                ArmadaSetupModel: DataView
            }
        }
    });

}(jQuery));
